# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from csv import QUOTE_MINIMAL


from duHast.Utilities.Objects.result import Result
from duHast.Revit.Family.family_types import (
    get_all_family_type_names, 
    delete_family_type, 
    create_family_type, 
    DEFAULT_CATALOGUE_REFERENCE_FAMILY_TYPE,
)
from duHast.Revit.Family.family_parameter_utils import set_family_parameter_value

from duHast.Utilities.files_io import file_exist
from duHast.Utilities.files_csv import read_csv_file, write_report_data_as_csv
from duHast.Utilities.Objects.file_encoding_bom import BOMValue


# if a default family type is created, these parameters will be reset to their default values
PARAMETERS_TO_RESET = [
    "HSL_AHFG_CODE",
    "HSL_AHFG_DESCRIPTION",
    "HSL_BUDGET_GROUP",
    "Description",
    "HSL_ID_TYPE",
    "Uniclass2015Code",
    "Uniclass2015Title",
    "Uniclass2015Version",
]

def update_type_catalogue_file(doc,  maintain_types_list):
    """
    Updates the type catalogue file for a family document by removing types that are not in the maintain list.

    :param doc: The family document to process.
    :type doc: :class:`Autodesk.Revit.DB.Document`
    :param maintain_types_list: A list of types to maintain, where each entry is a tuple (family_name, type_name).
    :type maintain_types_list: list of nested lists, each containing two strings: [family_name, type_name]
    :return: Result object containing the status and messages.
    :rtype: :class:`duHast.Utilities.Objects.result.Result`
    """

    return_value = Result()
    try:
        return_value.append_message(
            "Checking for type catalogue file for family '{}'...".format(doc.PathName)
        )

        # check if catalogue file exists
        catalogue_file_name =doc.PathName
        catalogue_file_name = catalogue_file_name[:-4] + ".txt"

        if not file_exist(catalogue_file_name):
            return_value.update_sep(
                True,
                "No type catalogue file found for family '{}'.".format(doc.Title),
            )
            return return_value
        
        return_value.append_message(
            "Found type catalogue file for family: {}".format(catalogue_file_name)
        )
        
        # read the type catalogue file
        read_result = read_csv_file(catalogue_file_name)

        if not read_result.status:
            return_value.update_sep(
                False,
                "Failed to read type catalogue file for family '{}': {}".format(doc.Title, read_result.message),
            )
            return return_value
        
        # the type name is the first entry in every data row, ignore the header row
        type_catalogue_data = read_result.result[1:]  # skip header row
        header_row = read_result.result[:1][0]  # keep the header row
        type_rows_to_maintain = []

        for row in type_catalogue_data:
            # check if the family name matches
            if row[0] in maintain_types_list:
                type_rows_to_maintain.append(row)

         # remove BOM from header row if it exists
        header_row[0] = "" 

        # write data back to file
        write_result = write_report_data_as_csv(
            file_name= catalogue_file_name,
            encoding="utf-16-le",
            bom=BOMValue.UTF_16_LITTLE_ENDIAN,
            header = header_row, # keep the header row from the read result
            data = type_rows_to_maintain,
            quoting= QUOTE_MINIMAL,  # use minimal quoting for CSV
        )
    
        return_value.update(write_result)

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to update type catalogue file in family '{}': {}".format(doc.Title, e),
        )
    return return_value


def get_names_of_types_to_keep(family_name, maintain_types_list):

    types_to_keep = []
    search_name = "{}.rfa".format(family_name)

    # loop over maintain types list and find the family name
    for family_type_row in maintain_types_list:
        if family_type_row[0] == search_name:
            types_to_keep.append(family_type_row[1])

    return types_to_keep


def create_default_type(doc, types_to_delete):
    """
    Creates a default family type in the family document if a catalogue file is in use...

    :param doc: The family document to process.
    :type doc: :class:`Autodesk.Revit.DB.Document`
    :return: Result object containing the status and messages.
    :rtype: :class:`duHast.Utilities.Objects.result.Result`
    """
    
    return_value = Result()
    try:

        # check if the default family type already exists
        if DEFAULT_CATALOGUE_REFERENCE_FAMILY_TYPE in types_to_delete:
            return_value.update_sep(
                True,
                "Default type '{}' already exists in family '{}'.".format(
                    DEFAULT_CATALOGUE_REFERENCE_FAMILY_TYPE, doc.Title
                ),
            )
            return return_value

        # create default catalogue type
        create_type_result = create_family_type(doc, DEFAULT_CATALOGUE_REFERENCE_FAMILY_TYPE)

        if not create_type_result.status:
            return_value.update_sep(
                False,
                "Failed to create default type in family '{}': {}".format(doc.Title, create_type_result.message),
            )
            return return_value
        
        # Get the FamilyManager
        family_manager = doc.FamilyManager

        # reset parameters
        for para_name in PARAMETERS_TO_RESET:
            # Get the parameter
            parameter = family_manager.get_Parameter(para_name)
            if parameter:
                set_result = set_family_parameter_value(doc, family_manager, parameter, DEFAULT_CATALOGUE_REFERENCE_FAMILY_TYPE)
                return_value.update(set_result)
            else: 
                return_value.append_message("Parameter not found: {}".format(para_name))

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to create default type in family '{}': {}".format(doc.Title, e),
        )
        return return_value
    return return_value


def delete_non_conforming_types(doc, maintain_types_list):
    """
    Delete non-conforming types in a family document based on a list of types to maintain.

    Note
    
    Types reported might only live in a type catalogue file. In that case the family may only contain a place holder type with name "Refer To Catalog File" or similar.

    :param doc: The family document to process.
    :type doc: :class:`Autodesk.Revit.DB.Document`
    :param maintain_types_list: A list of types to maintain, where each entry is a tuple (family_name, type_name).
    :type maintain_types_list: list of nested lists, each containing two strings: [family_name, type_name]

    :return: Result object containing the status and messages.
    :rtype: :class:`duHast.Utilities.Objects.result.Result`
    """
    return_value = Result()

    try:
        # get the family name
        family_name = doc.Title

        # build list of types to keep
        types_to_keep = get_names_of_types_to_keep(family_name, maintain_types_list)
        return_value.append_message(
            "Types to keep in family '{}': {}".format(family_name, len(types_to_keep))
        )

        all_type_names_result = get_all_family_type_names(doc)
        
        # check if we got any current type names
        if (not all_type_names_result.status):
            return_value.update_sep(
                False,
                "Failed to get all family type names: {}".format(all_type_names_result.message),
            )
            return return_value
        
        return_value.append_message(
            "Found {} types in family '{}'.".format(len(all_type_names_result.result), family_name)
        )
        
        # get all type names in family
        all_type_names = all_type_names_result.result

        # get list of names to delete
        types_to_delete = [name for name in all_type_names if name not in types_to_keep]

        if len(types_to_delete) == 0:
            return_value.update_sep(
                True,
                "No types to delete in family '{}'.".format(family_name),
            )
            return return_value
        
        # check if we are deleting all types in a family... if that is the case we will need to create a default type first
        if len(types_to_delete) == len(all_type_names):
            print("Deleting all types in family '{}'. Creating default type first.".format(family_name))
            create_default_result = create_default_type(doc, types_to_delete)
            if not create_default_result.status:
                return_value.update_sep(
                    False,
                    "Failed to create default type in family '{}': {}".format(family_name, create_default_result.message),
                )
                return return_value
            return_value.append_message(
                "Created default type in family '{}' successfully.".format(family_name)
            )

        # delete all other types
        for type_name in types_to_delete:
            # check if this is the default type
            if type_name == DEFAULT_CATALOGUE_REFERENCE_FAMILY_TYPE:
                return_value.append_message(
                    "Skipping deletion of default type '{}' in family '{}'.".format(type_name, family_name)
                )
                continue
            print("Deleting type '{}' in family '{}'...".format(type_name, family_name))
            delete_result = delete_family_type(doc, type_name)
            if not delete_result.status:
                return_value.append_message(
                    "Failed to delete type '{}': {}".format(type_name, delete_result.message)
                )
            else:
                return_value.append_message(
                    "Deleted type '{}' successfully.".format(type_name)
                )
        
        # fix up any type catalogue file
        update_type_catalogue_file_result = update_type_catalogue_file(doc, types_to_keep)
        return_value.update( update_type_catalogue_file_result)

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to delete non confirming types: {}".format(e),
        )
    
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit MEP systems names (ENG). 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# Duct types are split into three major families
#: Built in family name for oval ducting
DUCT_OVAL_FAMILY_NAME = "Oval Duct"
#: Built in family name for round ducting
DUCT_ROUND_FAMILY_NAME = "Round Duct"
#: Built in family name for rectangular ducting
DUCT_RECTANGULAR_FAMILY_NAME = "Rectangular Duct"

#: List of all Built in ducting family names
BUILTIN_DUCT_TYPE_FAMILY_NAMES = [
    DUCT_OVAL_FAMILY_NAME,
    DUCT_ROUND_FAMILY_NAME,
    DUCT_RECTANGULAR_FAMILY_NAME,
]

# flex duct types are split into two major families
#: Built in family name for rectangular flex ducting
FLEX_DUCT_REC_FAMILY_NAME = "Flex Duct Rectangular"
#: Built in family name for round flex ducting
FLEX_DUCT_ROUND_FAMILY_NAME = "Flex Duct Round"

#: List of all Built in flex ducting family names
BUILTIN_FLEX_DUCT_TYPE_FAMILY_NAMES = [
    FLEX_DUCT_REC_FAMILY_NAME,
    FLEX_DUCT_ROUND_FAMILY_NAME,
]

# conduits types are split into two major families
#: Built in family name for conduits with fittings
CONDUIT_WITH_FITTING_FAMILY_NAME = "Conduit with Fittings"
#: Built in family name for conduits without fittings
CONDUIT_WITHOUT_FITTING_FAMILY_NAME = "Conduit without Fittings"

#: List of all Built in conduit family names
BUILTIN_CONDUIT_TYPE_FAMILY_NAMES = [
    CONDUIT_WITH_FITTING_FAMILY_NAME,
    CONDUIT_WITHOUT_FITTING_FAMILY_NAME,
]

# cable tray types are split into two major families
#: Built in family name for cable tray with fittings
CABLE_TRAY_WITH_FITTING_FAMILY_NAME = "Cable Tray with Fittings"
#: Built in family name for cable tray without fittings
CABLET_RAY_WITHOUT_FITTING_FAMILY_NAME = "Cable Tray without Fittings"

#: List of all Built in cable tray family names
BUILTIN_CABLE_TRAY_TYPE_FAMILY_NAMES = [
    CABLE_TRAY_WITH_FITTING_FAMILY_NAME,
    CABLET_RAY_WITHOUT_FITTING_FAMILY_NAME,
]

# pipe types exist in one major families
#: Built in family name for pipes
PIPE_FAMILY_NAME = "Pipe Types"

#: List of all Built in pipe family names
BUILTIN_PIPE_TYPE_FAMILY_NAMES = [PIPE_FAMILY_NAME]



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit purging not used fill pattern styles using purge by delete helper functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.LinePattern.fill_patterns import (
    get_all_fill_pattern,
    get_used_pattern_ids_in_filled_region_types,
)
from duHast.Revit.Common.common import get_ids_from_element_collector
from duHast.Revit.Purge.purge_unused_by_delete import purge_unused_elements

from duHast.Utilities.Objects.result import Result


def get_fill_pattern_ids(doc, element_ids=None, element_ids_list_is_inclusive_filter=True):
    """
    Returns all fill pattern ids in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element_ids: optional list of fill pattern element ids
    :type element_ids: [Autodesk.Revit.DB.ElementId]
    :param element_ids_list_is_inclusive_filter: If true and element_ids list has values only those patterns will be purged if possible. If false and element_ids list has values any patterns in the list will not be purged.
    :type element_ids_list_is_inclusive_filter: bool
    
    :return: A list of all fill pattern ids in the model.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    fill_pattern_col = get_all_fill_pattern(doc)
    ids = get_ids_from_element_collector(fill_pattern_col)

    # reduce list of ids by removing ids used in filled region types
    # get all fill pattern ids used in filled region types
    used_pattern_ids = get_used_pattern_ids_in_filled_region_types(doc)
    # only return ids not used in filled region types
    ids_not_used_in_region_types = [id for id in ids if id not in used_pattern_ids]

    # check if further filtering is required
    if element_ids is None:
        return ids_not_used_in_region_types
    
    # apply filtering
    ids_filtered = []
    if element_ids_list_is_inclusive_filter:
        # only return element ids which are also present in the filter list
        for id_not_used in ids_not_used_in_region_types:
            if id_not_used in element_ids:
                ids_filtered.append(id_not_used)
    else:
        # only return element ids which are not present in the filter list
        for id_not_used in ids_not_used_in_region_types:
            if id_not_used not in element_ids:
                ids_filtered.append(id_not_used)
    return ids_filtered


def purge_fill_pattern_by_delete(doc, progress_callback=None, debug=False, element_ids=None, element_ids_list_is_inclusive_filter=True):
    """
    Purge fill pattern by delete.

    Note: This is a very slow process and can take a few hours to complete depending on the size of the model and the number of fill patterns.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param progress_callback: Callback to report progress.
    :type progress_callback: callable
    :param debug: Debug mode.
    :type debug: bool
    :param element_ids: optional list of fill pattern element ids
    :type element_ids: [Autodesk.Revit.DB.ElementId]
    :param element_ids_list_is_inclusive_filter: If true and element_ids list has values only those fill patterns will be purged if possible. If false and element_ids list has values any fill patterns in the list will not be purged.
    :type element_ids_list_is_inclusive_filter: bool

    :return: Result class instance.

        - .status True if unused fill pattern where deleted or nothing needed to be deleted. Otherwise False.
        - .message will contain deletion status.

    """

    return_value = Result()

    try:

        # set up element Id getter
        # make allowance for an ignore element id list
        def action(doc):
            result_action = get_fill_pattern_ids(
                doc=doc, element_ids=element_ids,
                element_ids_list_is_inclusive_filter=element_ids_list_is_inclusive_filter
            )
            return result_action
        
        # purge unused fill patterns
        purge_result = purge_unused_elements(
            doc=doc,
            element_id_getter=action,
            deleted_elements_modifier=None,
            modified_elements_modifier=None,
            progress_callback=progress_callback,
            debug=debug,
        )

        return_value.update(purge_result)
    except Exception as e:
        return_value.update_sep(
            False, "Error purging fill pattern by delete: {}".format(str(e))
        )

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A number of helper functions regarding advanced failure handling
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Helps to suppress warning dialogues in Revit.

"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

clr.AddReference("System.Core")
clr.ImportExtensions(System.Linq)

from System import EventHandler
from Autodesk.Revit.DB.Events import FailuresProcessingEventArgs
from Autodesk.Revit.DB import (
    FailureResolutionType,
    FailureSeverity,
    FailureProcessingResult,
)


from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.Objects.FailureHandlingConfiguration import (
    FailureHandlingConfig,
)
from duHast.Revit.Common.common import element_ids_to_semicolon_delimited_text


def get_failure_warning_report(failure, failure_definition):
    """
    Creates a string representation of the information attached to a failure

    :param failure: The failure to report
    :type failure: FailureMessage
    :param failure_definition: The failure definition
    :type failure_definition: FailureDefinition
    :return: The string representation of the failure information
    :rtype: str
    """

    result = Result()
    failure_message = []

    try:
        failure_severity = failure.GetSeverity()
        # Log basic failure information
        failure_message.append(
            "\t{} - {} - (GUID: {})".format(
                str(failure_severity),
                str(failure.GetDescriptionText()),
                str(failure.GetFailureDefinitionId().Guid),
            )
        )

        # Log the failing ids and additional elements (if required)
        if (
            failure_severity == FailureSeverity.Error
            or failure_severity == FailureSeverity.Warning
        ):
            failing_element_ids = failure.GetFailingElementIds()
            if len(failing_element_ids) > 0:
                failure_message.append("\n")
                failure_message.append(
                    "\tFailing element ids: {}".format(
                        element_ids_to_semicolon_delimited_text(failing_element_ids)
                    )
                )

            additional_elementIds = failure.GetAdditionalElementIds()
            if len(additional_elementIds) > 0:
                failure_message.append("\n")
                failure_message.append(
                    "\tAdditional element ids: {}".format(
                        element_ids_to_semicolon_delimited_text(additional_elementIds)
                    )
                )

        # If there are resolutions, log them
        if failure_severity == FailureSeverity.Error:
            if failure.HasResolutions():
                failure_message.append("\n\t" + "Applicable resolution types:\n")
                default_resolution_type = failure_definition.GetDefaultResolutionType()
                for (
                    resolution_type
                ) in failure_definition.GetApplicableResolutionTypes():
                    res_type = (
                        " (Default)"
                        if resolution_type == default_resolution_type
                        else ""
                    )
                    failure_message.append(
                        "\t\t{}{} - '{}'".format(
                            str(resolution_type),
                            res_type,
                            failure_definition.GetResolutionCaption(resolution_type),
                        )
                    )
            else:
                failure_message.append("\n")
                failure_message.append("\t" + "WARNING: no resolutions available")

        result.append_message ("\n".join(failure_message))
    except Exception as e:
        result.update_sep(False, "Failed to get failure warning report: {}".format(e))
    return result.message


def process_failures(failures_accessor, fail_config):
    """
    Process the failures encountered when transacting with a Revit document

    :param failures_accessor: The failures accessor to process
    :type failures_accessor: FailuresAccessor
    :return: The result of the failure processing
    :rtype: FailureProcessingResult
    """

    roll_back_on_warning = fail_config.roll_back_on_warning
    print_warnings = fail_config.print_warnings
    roll_back_on_error = fail_config.roll_back_on_error
    print_errors = fail_config.print_errors
    output = fail_config.output_function

    try:
        result = None
        # Get the failure registry
        doc = failures_accessor.GetDocument()
        app = doc.Application
        failure_registry = app.GetFailureDefinitionRegistry()
        failures = failures_accessor.GetFailureMessages()
        if failures.Any():
            if print_warnings or print_errors:
                output(
                    "Processing Revit document warnings / failures ({}):".format(
                        failures.Count
                    )
                )
            # Iterate through the failures
            for failure in failures:
                failure_definition = failure_registry.FindFailureDefinition(
                    failure.GetFailureDefinitionId()
                )
                failure_severity = failure.GetSeverity()
                # If warning and rolling back
                if failure_severity == FailureSeverity.Warning and roll_back_on_warning:
                    result = FailureProcessingResult.ProceedWithRollBack
                    if print_warnings:
                        output(get_failure_warning_report(failure, failure_definition))
                        output("\t" + "Rolled back warning")
                # If warning and not rolling back
                elif (
                    failure_severity == FailureSeverity.Warning
                    and not roll_back_on_warning
                ):
                    if print_warnings:
                        output(get_failure_warning_report(failure, failure_definition))
                    failures_accessor.DeleteWarning(failure)
                    result = FailureProcessingResult.ProceedWithCommit
                    if print_warnings:
                        output("\t" + "Deleted warning")

                # If error and rolling back
                elif failure_severity == FailureSeverity.Error and roll_back_on_error:
                    result = FailureProcessingResult.ProceedWithRollBack
                    if print_errors:
                        output(get_failure_warning_report(failure, failure_definition))
                        output("\t" + "Rolled back error")

                # If error and attempting resolution
                elif (
                    failure_severity == FailureSeverity.Error and not roll_back_on_error
                ):
                    if print_errors:
                        output(get_failure_warning_report(failure, failure_definition))

                    # Check if elements can be modified or if they are locked by another user
                    failure_guid = str(failure.GetFailureDefinitionId().Guid)
                    if failure_guid == "3b7fcaec-c01e-4c2e-819f-67ddd102ce1f":
                        output("\t" + "Cannot solve. Element locked by other user")
                        # locked element by other user. Can't do anything but roll back
                        result = FailureProcessingResult.ProceedWithRollBack
                    # If the failure has resolutions, attempt to resolve it
                    if failure.HasResolutions():
                        # If Unlock Constraints is a valid resolution type for the current failure, use it.
                        if failure.HasResolutionOfType(
                            FailureResolutionType.UnlockConstraints
                        ):
                            failure.SetCurrentResolutionType(
                                FailureResolutionType.UnlockConstraints
                            )
                        elif failure_definition.IsResolutionApplicable(
                            FailureResolutionType.UnlockConstraints
                        ):
                            # Revit is dumb. Unlock constraints cannot actually be used
                            output(
                                "\tWARNING: UnlockConstraints is not a valid resolution for this failure despite the definition reporting that it is an applicable resolution!"
                            )

                        # Will attempt to resolve the failure using the Default resolution type
                        output(
                            "\tAttempting to resolve error using resolution: {}".format(
                                failure.GetCurrentResolutionType()
                            )
                        )
                        failures_accessor.ResolveFailure(failure)
                        result = FailureProcessingResult.ProceedWithCommit
                    else:
                        output(
                            "\t"
                            + "WARNING: No resolutions available for this error. Rolling back"
                        )
                        result = FailureProcessingResult.ProceedWithRollBack

                # Every other scenario not caught by above
                else:
                    output("\t" + "No failure configuration match. Rolling back")
                    result = FailureProcessingResult.ProceedWithRollBack

        # No failures. Continue
        else:
            result = FailureProcessingResult.Continue
    except Exception as e:
        output("ERROR: the failure handler generated an error! \n{}".format(e))
        result = FailureProcessingResult.ProceedWithRollBack
    return result


def FailuresProcessingEventHandler(sender, args, fail_config):
    """
    Function to execute when Revit raises the FailuresProcessing event
    """

    failures_accessor = args.GetFailuresAccessor()
    failure_options = failures_accessor.GetFailureHandlingOptions()
    failure_options.SetForcedModalHandling(fail_config.set_forced_modal_handling)
    failure_options.SetClearAfterRollback(fail_config.set_clear_after_rollback)
    failures_accessor.SetFailureHandlingOptions(failure_options)
    failure_process_result = process_failures(failures_accessor, fail_config)
    args.SetProcessingResult(failure_process_result)
    return


def with_failures_processing_handler(app, action, fail_config=FailureHandlingConfig()):
    """
    Executes an action with Revit failure processing enabled by adding
    a handler to the FailuresProcessing event. This function will need
    to be wrapped in a transaction.

    :param app: The Revit application to execute the action with
    :type app: Application
    :param action: The action to perform
    :type action: function
    :return: The result of the action
    :rtype: object
    """
    result = None
    failure_processing_event_handler = EventHandler[FailuresProcessingEventArgs](
        lambda sender, args: FailuresProcessingEventHandler(sender, args, fail_config)
    )
    app.FailuresProcessing += failure_processing_event_handler
    try:
        result = action()
    finally:
        app.FailuresProcessing -= failure_processing_event_handler
    return result



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A number of functions around Revit independent tags.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import ElementId, Transaction, XYZ
from collections import namedtuple

from duHast.Utilities.Objects import result as res
from duHast.Revit.Common.transaction import in_transaction

from duHast.Revit.Common.revit_version import get_revit_version_number
from duHast.Utilities.files_json import read_json_data_from_file
from duHast.Revit.Annotation.Reporting.gen_annotations_instance_report_header import (
    TAG_ID,
    TAG_HEAD_LOCATION,
)
from duHast.Utilities.unit_conversion import convert_imperial_feet_to_metric_mm

# tuples containing move tag data
move_tag = namedtuple("move_tag", "tag new_location old_location")


def update_tag_location(doc, tag_data):
    """
    Moves all tags in dictionary to new location within one transaction

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param tag_data: A dictionary containing tag data of tag to be moved
    :type tag_data: {int:namedtuple move_tag}

    :return:
        Result class instance.

        - result.status. True if all tags where moved successfully, otherwise False.
        - result.message will contain ids of tags moved in format:'Moved tag: ' + id
        - result.result list of tags

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    def action():
        action_return_value = res.Result()
        try:
            for move_data in tag_data:
                try:
                    move_data.tag.Location.Move(
                        move_data.new_location - move_data.old_location
                    )
                    owner_view = doc.GetElement(move_data.tag.OwnerViewId)
                    action_return_value.update_sep(
                        True,
                        "Moved tag: {} on view: {} [view id: {}]".format(
                            move_data.tag.Id, owner_view.Name, owner_view.Id
                        ),
                    )
                    action_return_value.result.append(move_data.tag)
                except Exception as e:
                    action_return_value.update_sep(
                        False,
                        "Failed to move tag: {} with exception: {}".format(
                            move_data.tag.Id, e
                        ),
                    )
        except Exception as e:
            action_return_value.update_sep(
                False,
                "Failed to move tag with exception: {}".format(e),
            )
        return action_return_value

    # set up an transaction
    transaction = Transaction(doc, "Moving tags: {}".format(len(tag_data)))
    # execute transaction in transaction wrapper method
    result_moving_tags = in_transaction(transaction, action)
    return result_moving_tags


def update_tag_locations_from_report(doc, report_file_path, distance_threshold=50):
    """
    Reads a tag instance report and updates the location of tags in model with matching id if they have moved further then a given threshold distance.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param report_file_path: The fully qualified file path of the report file.
    :type report_file_path: str
    :param distance_threshold: The threshold distance, in mm, a tag has to move before it gets moved to a new location, defaults to 50. The distance is calculated by subtracting the current tag location from the one recorded in the report.
    :type distance_threshold: int, optional

    :return:
        Result class instance.

        - result.status. True if all tags where moved successfully, or no tags had to be moved, otherwise False.
        - result.message will contain ids of tags moved in format:'Moved tag: ' + id
        - result.result list of tags

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # read data from files
    read_result = read_json_data_from_file(report_file_path)
    if read_result.status is False:
        return_value.update_sep(
            False, "Failed to read report file: {}".format(read_result.message)
        )
        return return_value
    tag_data_list = read_result.result[0]
    
    tag_updates = []
    for tag_data in tag_data_list:
        # get the id of the tag
        id_data = tag_data[TAG_ID]
        try:
            # get the actual tag element in model
            tag_in_model = doc.GetElement(ElementId(id_data))
            # in case tag no longer exists
            if tag_in_model != None:
                head_location_data = tag_data[TAG_HEAD_LOCATION]
                head_location_data_as_xyz = XYZ(
                    head_location_data[0], head_location_data[1], head_location_data[2]
                )
                # check distance to current location
                tag_distance = abs(
                    convert_imperial_feet_to_metric_mm(
                        head_location_data_as_xyz.DistanceTo(
                            tag_in_model.TagHeadPosition
                        )
                    )
                )
                if tag_distance > distance_threshold:
                    # build a list of tags to move in one transaction
                    tag_updates.append(
                        move_tag(
                            tag=tag_in_model,
                            new_location=head_location_data_as_xyz,
                            old_location=tag_in_model.TagHeadPosition,
                        )
                    )
        except Exception as e:
            return_value.update_sep(
                False, "An exception occurred when gathering tag data: {}".format(e)
            )
    # check if any tags need updating their position:
    if len(tag_updates) > 0:
        return_value.update(update_tag_location(doc, tag_updates))
    else:
        return_value.update_sep(True, "No tags required moving")
    return return_value



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_xml import get_all_xml_files_from_directories

from duHast.Revit.Family.family_types_get_data_from_xml import get_family_type_data_from_library

from duHast.Revit.Family.LibraryCleanUp.Utility.directives_create import create_directives
from duHast.Revit.Family.LibraryCleanUp.Utility.directives_write_to_file import write_directives_to_file, write_maintain_list
from duHast.Revit.Family.LibraryCleanUp.Utility.directives_execute import execute_copy_directives_for_library_families
from duHast.Revit.Family.LibraryCleanUp.Utility.directives_checks import  check_directives
from duHast.Revit.Family.LibraryCleanUp.Utility.write_task_lists import write_task_lists

def get_family_data_from_file(libraryPath):
    """
    Extracts family type data from XML files in the specified library path.

    :return:
        Result class instance.

        - result.status: XML conversion status will be returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain which xml file was read and converted into family type data.
        - result.result will be [:class:`FamilyTypeDataStorageManager`]

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message
        - Reload.result will be an empty list

    :rtype: :class:`.Result`
    """
    xml_files =  get_all_xml_files_from_directories([libraryPath])

    if len(xml_files) == 0:
        return None
    
    family_data_result = get_family_type_data_from_library(xml_files)
    if (family_data_result.status is False):
        return None
    else:
        return family_data_result.result
    

def pre_process(library_path, output_path, task_list_directory_path, code_descriptor_path, output):
    """
    Pre-process function to prepare for the family type data extraction.
    
    This function is a placeholder for any pre-processing steps that may be needed before extracting family type data from the library.

    :param library_path: Path to the library containing family files to be renamed.
    :type library_path: str
    :param output_path: Path where the directives will be written and changed family files will be copied to.
    :type output_path: str
    :param task_list_directory_path: Path where the task lists for RBP will be written.
    :type task_list_directory_path: str
    :param code_descriptor_path: Path to the code descriptor file. (AUSHFG)
    :type code_descriptor_path: str

    :param output: Output object to write messages to.

    :return: Result object indicating the success or failure of the pre-processing step.
    :rtype: Result
    """
    return_value = Result()
    try:
        
        # do some logging to user
        output("Processing family data from library path: {}".format(library_path))

        # get type data from library as [:class:`.FamilyTypeDataStorageManager`]
        family_data = get_family_data_from_file(library_path)
        if (family_data is None):
            return_value.update_sep(False, "No family data found in the specified library path.")
            output("No family data found in the specified library path.")
            return return_value
        
        output("Families {} loaded from library path.".format(len(family_data)))
            
        # do some logging to user
        output("Creating directives for {} families.".format(len(family_data)))

        directives_result = None
        try:
            # built copy file directives / swap file directives
            directives_result = create_directives(family_data, output_path,  code_descriptor_path, output)
        except Exception as e:
            return_value.update_sep(
                False,
                "Failed to create directives with exception: {}".format(e),
            )
            output("Failed to create directives with exception: {}".format(e))
            return return_value

        if directives_result is None:
            return return_value
        elif directives_result and directives_result.status is False:
            return_value.update_sep(
                False,
                "Failed to create directives",
            )
            output("Failed to create directives: \n{}".format(directives_result.message))
            return return_value
        else:
            return_value.append_message(
                "Created {} directives.".format(len(directives_result.result))
            )
        
        # get directives
        copy_directives = directives_result.result[0]
        type_maintain_list = directives_result.result[1]
        swap_directives = directives_result.result[2]
        # check if fams with missing group codes were found
        families_with_missing_codes = None if len(directives_result.result) < 4 else directives_result.result[3]

        # sanity check directives
        check_result = check_directives(output_path, copy_directives, families_with_missing_codes)
        if check_result.status is False:
            return_value.update_sep(
                False,
                "Directives failed sanity check: {}".format(check_result.message),
            )
            output("Directives failed sanity check: {}".format(check_result.message))
            return return_value
        else:
            return_value.append_message(
                "Checked directives successfully."
            )

        # do some logging to user
        output("Created {} copy directives, {} type maintain directives and {} swap directives. ".format(
            len(copy_directives), len(type_maintain_list), len(swap_directives)))
        
        # write directives to file
        write_directives_result = write_directives_to_file (swap_directives, copy_directives, output_path)
        if write_directives_result.status is False:
            return_value.update_sep(
                False,
                "Failed to write directives to file",
            )
            return return_value
        else:
            return_value.append_message(
                "Wrote {} swap directives and {} type maintain directives to file.".format(
                    len(swap_directives), len(type_maintain_list))
            )
        
        output("Wrote directives to file.")

        # write type maintain list to file
        write_maintain_result = write_maintain_list(type_maintain_list, output_path)
        if write_maintain_result.status is False:
            return_value.update_sep(
                False,
                "Failed to write type maintain list to file",
            )
            output("Failed to write type maintain list to file: {}".format(write_maintain_result.message))
            return return_value
        else:
            return_value.append_message(
                "Wrote {} type maintain directives to file.".format(len(type_maintain_list))
            )

        output("Wrote maintain list to file.")

        # execute copy directives
        copy_result = execute_copy_directives_for_library_families(copy_directives)
        if copy_result.status is False:
            output("Failed to execute copy directives: {}".format(copy_result.message))
            return_value.update_sep(
                False,
                "Failed to execute copy directives",
            )
            return return_value
        else:
            return_value.append_message(
                "Executed {} copy directives.".format(len(copy_directives))
            )

        output("Executed copy directives.")

        # write task list to file
        write_task_lists_result = write_task_lists(
            family_directory=output_path, 
            task_list_directory = task_list_directory_path, 
            number_of_task_lists=3)
        
        output(write_task_lists_result)

        return return_value
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to get family data with exception: {}".format(e),
        )
        output("Failed to get family data with exception: {}".format(e))
    
    output("Finished!")
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions to rename family loaded family types in a project file or family file.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This helper function expect a folder containing rename directive files. For format of those files refer to module family_rename_files_utils.


"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# import clr
# import System

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Family.Utility import family_rename_types_utils as rFamRenameUtils
from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults  import NESTING_SEPARATOR
from duHast.Revit.Common import transaction as rTran

from Autodesk.Revit.DB import Element, Transaction


def _rename_loaded_family_types(doc, rename_directives, families, progress_callback=None):
    """
    Loops over families and if a match in rename directives is found will rename the family type accordingly.

    :param doc: The current family document.
    :type doc: Autodesk.Revit.DB.Document
    :param rename_directives: List of rename directives.
    :type rename_directives: [rename_directive]
    :param family_ids: dictionary of families where key is concatenated name + category and value is the family element.
    :type family_ids: {str:Autodesk.Revit.DB.Family}

    :return:
        Result class instance.

        - result.status. True if all families where renamed successfully, otherwise False.
        - result.message will contain each rename messages in format 'Renamed family from :' +current Name + ' to ' + newName.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message in format: 'Failed to rename family from :' + currentName + ' to ' + newName
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = Result()
    return_value.status = False
    rename_match_counter = 0

    # progress call back
    callback_counter = 1

    for rename_directive in rename_directives:

        # update progress
        if progress_callback != None:
            progress_callback.update(callback_counter, len(rename_directives))

        fam_key = "{}{}{}".format(rename_directive.name,NESTING_SEPARATOR,rename_directive.category)
        # check if match for rename directive
        if fam_key in families:

            # set counter indicating that at least one family was renamed
            rename_match_counter = rename_match_counter + 1

            # get the family
            family = families[fam_key]

            # reset flag
            found_type_name_match = False

            # get the type to rename
            for fam_type_id in family.GetFamilySymbolIds():
                family_type = doc.GetElement(fam_type_id)
                if Element.Name.GetValue(family_type)== rename_directive.old_type_name:
                    # flag that a match was found
                    found_type_name_match = True
                    # rename this family type
                    def action():
                        action_return_value = Result()
                        try:
                            # rename the type
                            family_type.Name = rename_directive.new_type_name
                                    
                            # give user feedback
                            if not found_type_name_match:
                                action_return_value.append_message("No match found for family [{}] of category [{}] vs directive category [{}] type: [{}] ".format(
                                    rename_directive.name,
                                    family.FamilyCategory.Name,
                                    rename_directive.category,
                                    rename_directive.old_type_name
                                ))
                            else:
                                action_return_value.update_sep(
                                    True,
                                    "Renamed family [{}] of category [{}] vs directive category [{}] from: [{}] to: [{}]".format(
                                        rename_directive.name,
                                        family.FamilyCategory.Name,
                                        rename_directive.category,
                                        rename_directive.old_type_name,
                                        rename_directive.new_type_name,
                                    ),
                                )
                        except Exception as e:
                            action_return_value.update_sep(
                                False,
                                "Failed to rename type of family [{}] of category [{}] vs directive category [{}] from: [{}] to: [{}]".format(
                                        rename_directive.name,
                                        family.FamilyCategory.Name,
                                        rename_directive.category,
                                        rename_directive.old_type_name,
                                        rename_directive.new_type_name,
                                    ),
                                )
                        return action_return_value

                    # start a transaction and rename the type
                    transaction = Transaction(doc, "Renaming: {} {}".format(rename_directive.name, rename_directive.old_type_name))
                    rename_result = rTran.in_transaction(transaction, action)
                    if rename_result.status:
                        # make sure that this returns true as soon as one family renamed successfully
                        return_value.status = True
                    # update messages
                    return_value.append_message(rename_result.message)
                    
                    # exit inner for loop
                    break
        else:
            # flag no match found
            return_value.append_message(
                "No match for rename directive for of name: {} and category: {} found.".format(
                    rename_directive.name, rename_directive.category
                )
            )

        # check for user cancel
        if progress_callback != None:
            if progress_callback.is_cancelled():
                return_value.append_message("User cancelled!")
                break
        # progress call back
        callback_counter += 1

    # check if anything got renamed at all
    if rename_match_counter == 0:
        return_value.append_message(
            "No match for rename directives found. Nothing was renamed."
        )
    return return_value


def rename_loaded_family_types(doc, directory_path):
    """
    Entry point for this module. Will read rename directives files in given directory and attempt to rename
    loaded family types accordingly.

    Note: for rename directive file structure refer to module family_rename_files_utils

    :param directory_path: Fully qualified directory path to where rename directive files are located.
    :type directory_path: str
    :return:
        Result class instance.

        - result.status. True if a single families was renamed successfully, otherwise False.
        - result.message will contain each rename messages in format 'Renamed family from :' +current Name + ' to ' + newName.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message in format: 'Failed to rename family from :' + currentName + ' to ' + newName
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = Result()
    # get directives from folder
    rename_directives_result = rFamRenameUtils.get_rename_directives(directory_path)
    # check if anything came back
    if rename_directives_result.status:
        rename_directives = rename_directives_result.result
        # get all family in file
        families = get_name_and_category_to_family_dict(doc)
        # check if any families are loaded
        if len(families) > 0:
            # rename files as per directives
            return_value = _rename_loaded_family_types(
                doc=doc, rename_directives=rename_directives, families=families
            )
        else:
            return_value.update_sep(True, "Mo loadable families in file.")
    else:
        return_value = rename_directives_result

    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions attempting to use .net library to write to file to avoid encoding errors. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import os
from System.Collections.Generic import List

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.Objects.file_encoding_bom import BOMValue as bom_value
from duHast.Revit.NetSupport.dll_names import FILE_WRAPPER 


# load the wrapper dll from the libs folder
# the dll is located in the libs folder of the extension, which is one level up from the current file's directory
current_directory = os.path.dirname(__file__)
parent_directory = os.path.dirname(current_directory)
dll_path = os.path.join(parent_directory, "lib", FILE_WRAPPER)
clr.AddReference(dll_path)

# import the WriteToFile class from the CSVHelperWrapper namespace
from duHastNet.FileIOWrapper import WriteToColumnBasedTextFile
from duHastNet.FileIOWrapper import BOMValue

def write_to_delimited_text_file(file_path, header, data, write_type="w", bom=None, delimiter=","):
    """
    Write data to a text file using the CSVHelperWrapper library.
    
    :param file_path: Path to the CSV file.
    :type file_path: str
    :param header: List of strings representing the header of the CSV file.
    :type header: List[str]
    :param data: List of list of strings representing the data to write to file.
    :type data: List[List[str]]
    :param write_type: Type of write operation. Default is "w" (write).
    :type write_type: str
    :param bom: BOM value. Default is None.
    :type bom: BOMValue or None
    :param delimiter: Delimiter used in the text file. Default is ",".
    :type delimiter: str
    
    :return:
        Result class instance.

        - result.status (bool) True if file was written without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will be an empty list.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
         result.result be an empty list
    :rtype: :class:`.Result`
    """
    
    return_value = Result()
    
    try:
        
        # do some type checking to ensure the inputs are valid
        if not isinstance(file_path, str):
            raise ValueError("file_path must be a string")
        if not isinstance(header, list) or not all(isinstance(item, str) for item in header):
            raise ValueError("header must be a list of strings")
        if not isinstance(data, list) or not all(isinstance(row, list) and all(isinstance(item, str) for item in row) for row in data):
            raise ValueError("data must be a list of lists of strings")
        
        
        # map bom values to the BOMValue enum
        if bom is not None:
            if bom == bom_value.UTF_8:
                bom = BOMValue.UTF8
            elif bom == bom_value.UTF_16_LITTLE_ENDIAN:
                bom = BOMValue.UTF_16_LITTLE_ENDIAN
            elif bom == bom_value.UTF_16_BIG_ENDIAN:
                bom = BOMValue.UTF_16_BIG_ENDIAN
            elif bom == bom_value.UTF_32_LITTLE_ENDIAN:
                bom = BOMValue.UTF_32_LITTLE_ENDIAN
            elif bom == bom_value.UTF_32_BIG_ENDIAN:
                bom = BOMValue.UTF_32_BIG_ENDIAN
            else:
                raise ValueError("Invalid BOM value. Must be one of the BOMValue enum values.")
        
        
        # convert the header and data to .NET List types
        # Convert to List[str]
        header_net = List[str]()  
        for item in header:
            # Add each item to the List[str]
            header_net.Add(item)  
        
        # Convert to List[List[str]]
        data_net = List[List[str]]() 
        for row in data:
            # Create a new List[str] for each row
            row_net = List[str]()  
            for item in row:
                # Add each item to the List[str]
                row_net.Add(item) 
                # Add the row to the List[List[str]]
            data_net.Add(row_net)  
        
        
        writer_net = WriteToColumnBasedTextFile()
        
        # Write to file using the WriteToFile class from the CSVHelperWrapper library
        result = writer_net.WriteToTextFile(
            file_path, 
            header_net, 
            data_net, 
            writeType=write_type, 
            bom=bom, 
            delimiter=delimiter,
        )
        
        # get all error messages from the WriteToFile class
        if not result:
            for message in writer_net.ErrorHistory:
                return_value.append_message(message)
            # set the overall status to False
            return_value.status = False
            # and get out
            return return_value
        
        # all went well, so set the status to True and append a message
        return_value.update_sep(result, "Text file written to {} with status: {}".format(file_path, result))
        return return_value

    except Exception as e:
        # handle any exceptions that occur during the writing process
        return_value.update_sep(False, "Error writing to text file: {}".format(e))
        return return_value




"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view sheets. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

from Autodesk.Revit.DB import (
    BuiltInCategory,
    BuiltInParameter,
    FilteredElementCollector,
    ViewSheet,
)


from duHast.Utilities import utility as util
from duHast.Revit.Views.Utility.sheet_parameters import get_sheet_number
from duHast.Revit.Common import parameter_get_utils as rParaGet

# required in lambda expressions!
clr.AddReference("System.Core")
import System

clr.ImportExtensions(System.Linq)


def get_sheets_by_filters(doc, view_rules=None):
    """
    Gets sheets matching filters provided
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view_rules: A set of rules. If sheet matches rule it will be returned. Defaults to None which will return all sheets.
    :type view_rules: array in format [parameter name, condition test method, value to test against], optional

    :return: Views matching filter
    :rtype: list of Autodesk.Revit.DB.View
    """

    collector_views = FilteredElementCollector(doc).OfClass(ViewSheet)
    views = []
    for v in collector_views:
        # if no filter rules applied return al sheets
        if view_rules is not None:
            paras = v.GetOrderedParameters()
            rule_match = True
            for para_name, paraCondition, conditionValue in view_rules:
                for p in paras:
                    if p.Definition.Name == para_name:
                        rule_match = rule_match and rParaGet.check_parameter_value(
                            p, paraCondition, conditionValue
                        )
            if rule_match == True:
                # delete view
                views.append(v)
        else:
            views.append(v)
    return views


def get_all_sheets(doc):
    """
    Gets all sheets in a model
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: list of sheet views
    :rtype: list of Autodesk.Revit.DB.View
    """

    collector_views = FilteredElementCollector(doc).OfClass(ViewSheet)
    return collector_views


def get_sheet_rev_by_sheet_number(doc, sheet_number):  # type # type: str
    """
    Returns the revision of a sheet identified by its number. Default value is '-'.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param sheet_number: The number of the sheet of which the revision is to be returned.
    :type sheet_number: str
    :raise: Any exception will need to be managed by the function caller.

    :return: The sheet's current revision value. If no matching sheet is found, '-' is returned.
    :rtype: str
    """

    rev_value = "-"
    collector = (
        FilteredElementCollector(doc)
        .OfClass(ViewSheet)
        .Where(lambda e: e.SheetNumber == sheet_number)
    )  # type
    results = collector.ToList()
    if len(results) > 0:
        sheet = results[0]
        rev_p = sheet.get_Parameter(BuiltInParameter.SHEET_CURRENT_REVISION)
        rev_value = rev_p.AsString()
    return rev_value


def get_sheet_rev_by_sheet_name(doc, sheet_name):  # type # type: str
    """
    Returns the revision of a sheet identified by its name. Default value is '-'.

    Since multiple sheets can have the same name, it will return the revision of the first sheet matching the name.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param sheet_name: The name of the sheet of which the revision is to be returned.
    :type sheet_name: str
    :raise: Any exception will need to be managed by the function caller.

    :return: The sheet's current revision value. If no matching sheet is found, '-' is returned.
    :rtype: str
    """

    rev_value = "-"
    collector = (
        FilteredElementCollector(doc)
        .OfClass(ViewSheet)
        .Where(lambda e: e.Name == sheet_name)
    )  # type
    results = collector.ToList()
    if len(results) > 0:
        sheet = results[0]
        rev_p = sheet.get_Parameter(BuiltInParameter.SHEET_CURRENT_REVISION)
        rev_value = util.pad_single_digit_numeric_string(rev_p.AsString())
    return rev_value


def get_sheet_num_to_elem_dict(rvt_doc):
    """
    Get a dictionary of sheet numbers and sheets
    :param rvt_doc: The Revit document to get the sheets from
    :type rvt_doc: Document
    :return: A dictionary of sheet numbers and sheets
    :rtype: dict
    """
    num_elem_dict = {}
    all_sheets = get_all_sheets(rvt_doc)

    for sheet in all_sheets:
        sht_num = get_sheet_number(sheet)
        num_elem_dict[sht_num] = sheet

    return num_elem_dict


def get_title_block_from_sheet(doc, sheet):
    """
    Returns the first title block located on a sheet or None if there is none placed.

    Args:
        doc (Autodesk.Revit.DB.Document): Current Revit model document.
        sheet (Autodesk.Revit.DB.ViewSheet):The sheet of which to return the title block from

    Returns:
        Autodesk.Revit.DB.FamilyInstance: The first title block instance on a sheet or None.
    """
    
    collector = (
        FilteredElementCollector(doc, sheet.Id)
        .OfCategory(BuiltInCategory.OST_TitleBlocks)
        .WhereElementIsNotElementType()
    )
    for element in collector:
        return element
    
    # no title block?
    return None



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions around families used in system types.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb
from duHast.Revit.Common import common as com
from duHast.Revit.MEP_Systems.Utility.MergeLists import merge_into_unique_list
from duHast.Revit.MEP_Systems.Utility.RevitMEPSystemCategories import (
    ROUTING_PREF_RULE_GROUP_TYPES,
)

"""
Properties of system types which can use symbols: (note: RoutingPreferenceManager may contain multiple rules per connection type!)

- Cross
- Elbow
- MultiShapeTransition
- Tap
- Tee
- Transition
- Union

"""


def get_unique_ids_of_used_symbols_from_system_type_id(doc, system_type_id):
    """
    Gets list of unique symbol ids used in a single system type property.
    List can be empty if an exception during processing occurred.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param system_type_id: MEP system type id (pipe, conduit, duct, cable tray)
    :type system_type_id: Autodesk.Revit.DB.ElementId
    :return: List of unique ids representing family symbols used in a system.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = []
    el = doc.GetElement(system_type_id)
    try:
        unfiltered_elements = [
            el.Cross,
            el.Elbow,
            el.MultiShapeTransition,
            el.Tap,
            el.Tee,
            el.Transition,
            el.Union,
        ]
        for unfiltered_element in unfiltered_elements:
            if unfiltered_element != None:
                if (
                    unfiltered_element.Id != rdb.ElementId.InvalidElementId
                    and unfiltered_element.Id not in ids
                ):
                    ids.append(unfiltered_element.Id)
        # check if there is a RoutingPreferenceManager object...it may have some more symbols in its rules
        if el.RoutingPreferenceManager != None:
            # routing manager got a list RoutingReferenceRule objects
            # each of those got a 	MEPPartId property which is what we are after
            rpm = el.RoutingPreferenceManager
            for group in ROUTING_PREF_RULE_GROUP_TYPES:
                # loop over RoutingPreferenceRuleGroupTypes!
                number_of_rules = rpm.GetNumberOfRules(group)
                for i in range(number_of_rules):
                    rule = rpm.GetRule(group, i)
                    if rule.MEPPartId not in ids:
                        ids.append(rule.MEPPartId)

    except Exception as ex:
        print(
            "System type id: {} get used symbol ids threw exception: {}".format(
                system_type_id, ex
            )
        )
    return ids


def get_unique_ids_of_used_symbols_from_system_type_ids(doc, system_type_ids):
    """
    Gets a list of unique symbol ids used in these MEP system type properties:
    - Cross
    - Elbow
    - MultiShapeTransition
    - Tap
    - Tee
    - Transition
    - Union
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param system_type_ids: List of MEP system type id belonging to pipe, conduit, duct or cable tray.
    :type system_type_ids: List Autodesk.Revit.DB.ElementId
    :return: List of unique ids representing family symbols used in mep systems.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = []
    for system_type_id in system_type_ids:
        ids_unfiltered = get_unique_ids_of_used_symbols_from_system_type_id(
            doc, system_type_id
        )
        ids = merge_into_unique_list(ids, ids_unfiltered)
    return ids


# --------------------------------------- symbols available in model -------------------------------


def get_symbol_ids_of_mep_system_types(doc, category_list, system_type_name):
    """
    Gets list of symbol ids belonging to provided categories loaded in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param category_list: List of built in categories to filter symbols by.
    :type category_list: list Autodesk.Revit.DB.BuiltInCategory
    :param system_type_name: Used in exception message to identify the mep system
    :type system_type_name: str
    :return: List of ids representing family symbols.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    ids = []
    try:
        multi_cat_filter = rdb.ElementMulticategoryFilter(category_list)
        col = (
            rdb.FilteredElementCollector(doc)
            .OfClass(rdb.FamilySymbol)
            .WherePasses(multi_cat_filter)
        )
        ids = com.get_ids_from_element_collector(col)
    except Exception as ex:
        print(system_type_name + " threw exception: " + str(ex))
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Interface for family processing class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import System
from Autodesk.Revit.DB import ElementId
import json
from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Revit.Family.Data.Objects.ifamily_data_storage import IFamilyDataStorage
from duHast.Utilities.Objects import result as res
from duHast.Utilities.Objects import base
from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults import (
    NESTING_SEPARATOR,
)


class IFamilyProcessor(base.Base):

    data_type_header = "Data Type"  # Data type header

    def __init__(self, data_type=None, pre_actions=None, post_actions=None, **kwargs):

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(IFamilyProcessor, self).__init__(**kwargs)

        self.data = []

        # check if data type is set
        if data_type != None:
            self.data_type = data_type
        else:
            self.data_type = "not declared"

        self.pre_actions = pre_actions
        self.post_actions = post_actions

    # -------------------------------------- utility ----------------------

    def _update_data(
        self,
        processor,
        identify_by_this_property_name,
        identify_by_this_property_value,
        update_by_property_name,
        updated_to_this_property_value,
    ):
        update_status = processor.update_Data(
            identify_by_this_property_name,
            identify_by_this_property_value,
            update_by_property_name,
            updated_to_this_property_value,
        )
        return update_status

    def _find_first_root_family_data(self):
        """
        Finds the data instance which contains the processed the root family.

        :return: data instance
        :rtype: IFamilyData
        """

        for data in self.data:
            if isinstance(data, IFamData.IFamilyData):

                data_stored = data.get_data()
                if isinstance(data_stored, list):
                    for storage_entry in data_stored:
                        if isinstance(storage_entry, IFamilyDataStorage):
                            if NESTING_SEPARATOR not in storage_entry.root_name_path:
                                return data
                        else:
                            raise ValueError(
                                "Data must be of type IFamilyDataStorage but is of type: [{}]".format(
                                    type(storage_entry)
                                )
                            )
                else:
                    raise ValueError(
                        "Data must be of type list but is of type: [{}]".format(
                            type(data_stored)
                        )
                    )
            else:
                raise ValueError(
                    "Data must be of type IFamilyData but is of type: [{}]".format(
                        type(data)
                    )
                )

    def _find_root_family_data(self):
        """
        Returns all data instances from root families (top most in tree) .

        :return: List of data instances.
        :rtype: [IFamilyData]
        """
        family_data = []
        for data in self.data:
            if isinstance(data, IFamData.IFamilyData):
                if NESTING_SEPARATOR not in data.root_path:
                    family_data.append(data)
            else:
                raise ValueError(
                    "Data must be of type IFamilyData but is of type: [{}]".format(
                        type(data)
                    )
                )
        return family_data

    def _find_nested_families_data(self):
        """
        Returns all data instances from nested families from each processor instances.

        :return: List of dictionaries.
        :rtype: [IFamilyData]
        """

        nested_family_data = []

        for data in self.data:
            if isinstance(data, IFamData.IFamilyData):
                if NESTING_SEPARATOR in data.root_path:
                    nested_family_data.append(data)
            else:
                raise ValueError(
                    "Data must be of type IFamilyData but is of type: [{}]".format(
                        type(data)
                    )
                )

        return nested_family_data

    # -------------------------------------- pre process actions ----------------------

    def pre_process_actions(self, doc):
        """
        Actions any pre processing before family data will be collected.

        :param doc: The family document.
        :type doc: Autodesk.Revit.DB.Document

        :return: _description_
        :rtype: _type_
        """

        return_value = res.Result()
        if self.pre_actions != None:
            for pre_action in self.pre_actions:
                result_action = pre_action(doc)
                return_value.update(result_action)
        return return_value

    # -------------------------------------- process actions ----------------------

    def process(self, doc, root_path, root_category_path):
        """
        Gather data on the root family and any nested families

        :param doc: The family document. 
        :type doc: Autodesk.Revit.DB.Document

        :param root_path: The path of the nested family in a tree: rootFamilyName :: nestedFamilyNameOne :: nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type root_path: str
        :param root_category_path: The path of the nested family category in a tree: rootFamilyCategory :: nestedFamilyOneCategory :: nestedFamilyTwoCategory\
            This includes the actual family category as the last node.
        :type root_category_path: str
        """

        pass

    # -------------------------------------- post process actions ----------------------

    def post_process_actions(self, doc):
        """
        Actions any post processing after family data has been collected.

        :param doc: The family document.
        :type doc: Autodesk.Revit.DB.Document
        """

        return_value = res.Result()
        if self.post_actions != None:
            for post_action in self.post_actions:
                result_action = post_action(doc)
                return_value.update(result_action)
        return return_value

    # -------------------------------------- get data ----------------------

    def get_data(self):
        """
        Returns list of data storage instances.

        :return: List of iFamilyStorage instances.
        :rtype: [iFamilyStorage]
        """

        data_out = []
        for data in self.data:
            if isinstance(data, IFamData.IFamilyData):
                data_stored = data.get_data()
                if isinstance(data_stored, list):
                    for storage_entry in data_stored:
                        if isinstance(storage_entry, IFamilyDataStorage):
                            data_out.append(storage_entry)
                        else:
                            raise ValueError(
                                "Data must be of type IFamilyDataStorage but is of type: [{}]".format(
                                    type(storage_entry)
                                )
                            )
                else:
                    raise ValueError(
                        "Data must be of type list but is of type: [{}]".format(
                            type(data_stored)
                        )
                    )
            else:
                raise ValueError(
                    "Data must be of type IFamilyData but is of type: [{}]".format(
                        type(data)
                    )
                )
        return data_out

    def get_data_json(self):
        """
        Returns data objects as JSON formatted strings.

        :return: JSON formatted string.
        :rtype: str
        """

        out_value = ""
        flattened_storage_data = self.get_data()
        for storage in flattened_storage_data:
            json_object = storage.to_json()
            out_value = out_value + "\n" + json_object
        return out_value

    def get_data_string_list(self):
        """
        Returns data storage objects as nested list of strings, where each inner list represents the values of a data storage object.

        - Strings are UTF 8 encoded


        :return: list of list of strings.
        :rtype: [str]
        """
        out_value = []

        flattened_storage_data = self.get_data()
        # print("flattened_storage_data: ", flattened_storage_data)
        for storage in flattened_storage_data:
            out_value.append(storage.get_data_values_as_list_of_strings())
            # print("out_value: ", out_value)

        # print("out_value: ", out_value)
        return out_value

    def get_data_headers(self):
        """
        Returns the headers of the data storage objects.

        :return: list of strings.
        :rtype: [str]
        """

        # get the first storage object
        if len(self.data) > 0:
            data_instance = self.data[0]
            return data_instance.get_property_names()
        else:
            return []



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A number of functions Excel file operations.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

clr.AddReference("Microsoft.Office.Interop.Excel")
import Microsoft.Office.Interop.Excel as Excel

import os

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_io import (
    get_directory_path_from_file_path,
    file_delete,
    get_file_name_without_ext,
)
from duHast.Utilities.files_csv import read_csv_file
from duHast.Utilities.benchmarking import add_measure_time


def get_excel_tab_names(file_path):
    """
    Returns a list of tab names (work sheet names) in the Excel file.

    :param file_path: The path to the Excel file.
    :type file_path: str
    :return: A result object with the result set to a list of tab names if successful, otherwise an error message.
    :rtype: Result
    """

    return_value = Result()
    excel = None
    wb = None
    try:
        excel = Excel.ApplicationClass()
        excel.Visible = False
        excel.DisplayAlerts = False
        # refer to this website for arg docs:https://learn.microsoft.com/en-us/office/vba/api/Excel.Workbooks.Open
        wb = excel.Workbooks.Open(
            file_path,  # excel doc file path
            0,  # update links ( 0 No )
            True,  # read only
            5,  # format delimeter ( 5 = nothing )
            "",  # password to open  a file (not required)
            "",  # password to write to a write protected  file (not required)
            True,  # ignore read only recommended (False = don't ignore, True = ignore)
            Excel.XlPlatform.xlWindows,  # file origin (xlWindows = Microsoft Windows)
            "\t",  # delimiter (tab) optional
            False,  # editable (False = read only, True = editable)
            False,  # notify (False = don't notify, True = notify)
            0,  # converter (0 = don't convert, 1 = convert)
            False,  # add to MRU (False = don't add, True = add)
            True,  # local (False = don't local, True = local)
            False,  # corrupt load (False = don't load, True = load)
        )
        tab_names = [ws.Name for ws in wb.Worksheets]
        return_value.update_sep(True, "Successfully read Excel file.")
        return_value.result = tab_names
    except Exception as e:
        return_value.update_sep(False, "Failed to read Excel file. Error: {}".format(e))
    finally:
        if wb:
            wb.Close(False)
        if excel:
            excel.Quit()

    return return_value


def save_excel_file_as_csv(file_path_excel, file_path_csv, tab_name=None):
    """
    Saves an Excel file as a CSV file.

    :param file_path_excel: The path to the Excel file.
    :type file_path_excel: str
    :param file_path_csv: The path to the CSV file.
    :type file_path_csv: str
    :param tab_name: The name of the tab ( work sheet ) in the Excel file to save as CSV. If None, the first tab is saved.
    :type tab_name: str
    :return: A result object with the result set to True if successful, otherwise an error message.
    :rtype: Result
    """

    return_value = Result()
    excel = None
    wb = None
    try:
        excel = Excel.ApplicationClass()
        excel.Visible = False
        excel.DisplayAlerts = False
        # refer to this website for arg docs:https://learn.microsoft.com/en-us/office/vba/api/Excel.Workbooks.Open
        wb = excel.Workbooks.Open(
            file_path_excel,  # excel doc file path
            0,  # update links ( 0 No )
            True,  # read only
            5,  # format delimeter ( 5 = nothing )
            "",  # password to open  a file (not required)
            "",  # password to write to a write protected  file (not required)
            True,  # ignore read only recommended (False = don't ignore, True = ignore)
            Excel.XlPlatform.xlWindows,  # file origin (xlWindows = Microsoft Windows)
            "\t",  # delimiter (tab) optional
            False,  # editable (False = read only, True = editable)
            False,  # notify (False = don't notify, True = notify)
            0,  # converter (0 = don't convert, 1 = convert)
            False,  # add to MRU (False = don't add, True = add)
            True,  # local (False = don't local, True = local)
            False,  # corrupt load (False = don't load, True = load)
        )
        if tab_name:
            ws = wb.Worksheets[tab_name]
            ws.Activate()
        else:
            ws = wb.Worksheets[1]
            ws.Activate()

        wb.SaveAs(file_path_csv, Excel.XlFileFormat.xlCSV)
        return_value.update_sep(True, "Successfully saved Excel file as CSV.")
    except Exception as e:
        return_value.update_sep(
            False, "Failed to save Excel file as CSV. Error: {}".format(e)
        )
    finally:
        if wb:
            wb.Close(False)
        if excel:
            excel.Quit()
    return return_value


@add_measure_time
def read_excel_file(file_path, excel_tab_name=None):
    """
    Reads the content of an Excel file into a 2D array.

    Note: this function is rather slow, so use it with caution.

    :param file_path: The path to the Excel file.
    :type file_path: str
    :param excel_tab_name: The name of the tab ( work sheet ) in the Excel file to read from. If None, the first tab is read.
    :type excel_tab_name: str
    :return: A result object with the result set to a 2D array if successful, otherwise an error message.
    :rtype: Result
    """

    return_value = Result()
    excel = None
    wb = None
    try:

        @add_measure_time
        def init_excel():
            excel = Excel.ApplicationClass()
            excel.Visible = False
            excel.DisplayAlerts = False
            return excel

        excel = init_excel()

        @add_measure_time
        def open_workbook():
            # refer to this website for arg docs:https://learn.microsoft.com/en-us/office/vba/api/Excel.Workbooks.Open
            wb = excel.Workbooks.Open(
                file_path,  # excel doc file path
                0,  # update links ( 0 No )
                True,  # read only
                5,  # format delimeter ( 5 = nothing )
                "",  # password to open  a file (not required)
                "",  # password to write to a write protected  file (not required)
                True,  # ignore read only recommended (False = don't ignore, True = ignore)
                Excel.XlPlatform.xlWindows,  # file origin (xlWindows = Microsoft Windows)
                "\t",  # delimiter (tab) optional
                False,  # editable (False = read only, True = editable)
                False,  # notify (False = don't notify, True = notify)
                0,  # converter (0 = don't convert, 1 = convert)
                False,  # add to MRU (False = don't add, True = add)
                True,  # local (False = don't local, True = local)
                False,  # corrupt load (False = don't load, True = load)
            )
            return wb

        wb = open_workbook()

        if excel_tab_name:
            ws = wb.Worksheets[excel_tab_name]
        else:
            ws = wb.Worksheets[1]

        # get the used range of the worksheet
        used_range = ws.UsedRange
        rows_count = used_range.Rows.Count
        cols_count = used_range.Columns.Count

        # read the data into a 2D array
        used_data = []

        @add_measure_time
        def iterate_through_cells():
            try:
                # Read the entire range at once - much faster than cell by cell
                range_values = used_range.Value2

                if range_values is None:
                    # Empty range - fill with None values
                    for _ in range(rows_count):
                        used_data.append([None] * cols_count)
                elif rows_count == 1 and cols_count == 1:
                    # Single cell case - Value2 returns a single value
                    used_data.append([range_values])
                elif rows_count == 1:
                    # Single row case - convert to list
                    if hasattr(range_values, "__iter__"):
                        row_data = [val for val in range_values]
                    else:
                        row_data = [range_values]
                    used_data.append(row_data)
                elif cols_count == 1:
                    # Single column case - each value becomes a single-item list
                    if hasattr(range_values, "__iter__"):
                        for val in range_values:
                            used_data.append([val])
                    else:
                        used_data.append([range_values])
                else:
                    # Multi-row, multi-column case
                    # Convert COM array to Python list using list comprehension
                    for row in range_values:
                        if hasattr(row, "__iter__"):
                            used_data.append([val for val in row])
                        else:
                            used_data.append(row)

            except Exception as e:
                return_value.update_sep(
                    False, "Failed to read Excel file. Error: {}".format(e)
                )

        iterate_through_cells()

        @add_measure_time
        def pair_values(in_list):
            """
            Break up the list into chunks based on the column count from Excel.
            Each chunk will contain cols_count elements.
            """
            paired = []
            for i in range(0, len(in_list), cols_count):
                chunk = in_list[i : i + cols_count]
                # Pad with None if the chunk is shorter than expected
                while len(chunk) < cols_count:
                    chunk.append(None)
                paired.append(chunk)
            return paired

        return_value.update_sep(True, "Successfully read Excel file.")
        return_value.result = pair_values(used_data)
    except Exception as e:
        return_value.update_sep(False, "Failed to read Excel file. Error: {}".format(e))
    finally:
        if wb:
            wb.Close(False)
        if excel:
            excel.Quit()

    return return_value


def read_excel_file_fast(file_path, excel_tab_name=None):
    """
    Reads the content of an Excel file into a 2D array.

    Note: this function is faster than read_excel_file
    It will save the excel file as a csv file first and then read its content.

    :param file_path: The path to the Excel file.
    :type file_path: str
    :param excel_tab_name: The name of the tab ( work sheet ) in the Excel file to read from. If None, the first tab is read.
    :type excel_tab_name: str
    :return: A result object with the result set to a 2D array if successful, otherwise an error message.
    :rtype: Result
    """

    return_value = Result()
    temp_csv_path = None
    try:
        temp_csv_path = os.path.join(
            get_directory_path_from_file_path(file_path), "temp.csv"
        )

        conversion_result = save_excel_file_as_csv(
            file_path, temp_csv_path, excel_tab_name
        )
        if not conversion_result.status:
            raise ValueError(conversion_result.message)
        return_value.update(conversion_result)

        # attempt read the csv file
        data_result = read_csv_file(temp_csv_path)
        if data_result.status is False:
            raise ValueError(data_result.message)
        data = data_result.result

        return_value.append_message(
            "Successfully read temp csv file. Number of rows read: {}".format(len(data))
        )
        return_value.result = data

    except Exception as e:
        return_value.update_sep(False, "Failed to read Excel file. Error: {}".format(e))

    finally:
        if temp_csv_path:
            delete_flag = file_delete(temp_csv_path)
            return_value.append_message(
                "Deleted temp csv file with status [{}]".format(delete_flag)
            )

    return return_value


def save_csv_as_excel(file_path):
    """
    Convertes as csv to .xslx in the same directory.

    :param file_path: The path to the csv file.
    :type file_path: str

    :return: A result object with the result status set to true if successful, otherwise an error message.
    :rtype: Result
    """

    return_value = Result()
    excel = None
    file_name_new = ""
    try:
        excel = Excel.ApplicationClass()
        excel.Visible = False
        excel.DisplayAlerts = False
        # refer to this website for arg docs:https://learn.microsoft.com/en-us/office/vba/api/Excel.Workbooks.Open
        # open excel and the csv
        wb = excel.Workbooks.Open(
            file_path,  # excel doc file path
            0,  # update links ( 0 No )
            True,  # read only
            6,  # format delimeter ( 6 = CSV)
            "",  # password to open a file (not required)
            "",  # password to write to a write protected file (not required)
            True,  # ignore read only recommended (False = don't ignore, True = ignore)
            Excel.XlPlatform.xlWindows,  # file origin (xlWindows = Microsoft Windows)
            ",",  # delimiter (comma) optional
            False,  # editable (False = read only, True = editable)
            False,  # notify (False = don't notify, True = notify)
            0,  # converter (0 = don't convert, 1 = convert)
            False,  # add to MRU (False = don't add, True = add)
            True,  # local (False = don't local, True = local)
            False,  # corrupt load (False = don't load, True = load)
        )
        # build new file name:
        output_dir = get_directory_path_from_file_path(file_path)
        file_name_part = get_file_name_without_ext(file_path=file_path)
        file_name_new = os.path.join(output_dir, file_name_part + ".xlsx")

        # save:
        save_status = wb.SaveAs(file_name_new, Excel.XlFileFormat.xlWorkbookDefault)

        # save the file
        wb.Close(False)

    except Exception as e:
        return_value.update_sep(
            False, "Failed to save csv with exception: {}".format(e)
        )

    if excel is not None:
        excel.Quit()

    return_value.update_sep(
        True, "Successfully saved Excel file: {}.".format(file_name_new)
    )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit grids helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

import System

# import common library modules
from duHast.Revit.Common import common as com
from duHast.Revit.Common import parameter_get_utils as rParaGet
from duHast.Revit.Common.Geometry.points import get_point_as_string

# import Autodesk
from Autodesk.Revit.DB import (
    BuiltInCategory,
    DatumExtentType,
    Element,
    ElementCategoryFilter,
    ElementId,
    FamilySymbol,
    Grid,
    Line,
    FilteredElementCollector,
    View,
)


def get_grids_in_model(doc):
    """
    Get all grids in model

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A collector with all grids in model.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = FilteredElementCollector(doc).OfClass(Grid)
    return collector


def get_multi_segmented_grids(doc):
    """
    Get all multi segmented grids in model

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A collector with all multi segmented grids in model.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = FilteredElementCollector(doc).OfCategory(BuiltInCategory.OST_GridChains)
    return collector


def get_grids_in_view(doc, view):
    """
    Get all grids in a view

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view: The view to get grids from.
    :type view: Autodesk.Revit.DB.View
    :return: A collector with all grids in view.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = FilteredElementCollector(doc, view.Id).OfClass(Grid)
    return collector


def get_multi_segmented_grids_in_view(doc, view):
    """
    Get all multi segmented grids in a view

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view: The view to get grids from.
    :type view: Autodesk.Revit.DB.View
    :return: A collector with all multi segmented grids in view.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = FilteredElementCollector(doc, view.Id).OfCategory(
        BuiltInCategory.OST_GridChains
    )
    return collector


def get_all_segment_grid_ids(multi_segment_grids):
    """
    Gets a unique list of grids id's making up the multi segment grids past in.

    :param multi_segment_grids: List of multi segment grids
    :type multi_segment_grids: [_type_]

    :return: List of individual grid ids.
    :rtype:[Autodesk.Revit.DB.ElementId]
    """

    ids = []
    for g in multi_segment_grids:
        sIds = g.GetGridIds()
        for sId in sIds:
            if sId not in ids:
                ids.append(sId)
    return ids


def get_linear_grids_in_model_by_view(doc, view):
    """
    Returns only linear grid visible in view.

    Will exclude multi segment grids and arc based grids.

    :param doc: Current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view: The view.
    :type view: Autodesk.Revit.DB.View

    :return: List of linear grids.
    :rtype: [Autodesk.Revit.DB.Grid]
    """

    # do some type checking
    if isinstance(view, View) == False:
        raise TypeError(
            "view needs to be of type Autodesk.Revit.DB.View. Got {} instead.".format(
                type(view)
            )
        )

    # get all grids in view
    grids = get_grids_in_view(doc, view).ToElements()
    # get all multi segment grids in view
    grids_multi = get_multi_segmented_grids_in_view(doc, view).ToElements()
    # get the individual grid id's making up the multi segment grid
    segment_grid_ids = get_all_segment_grid_ids(grids_multi)
    # set up placeholder to be returned
    grids_filtered = []
    # loop over grids in view and check:
    for g in grids:
        # is this grid part of a multi segment grid?
        if g.Id not in segment_grid_ids:
            # get the curve defining the grid
            curve_grid = g.Curve
            # check if the curve is of type line
            if type(curve_grid) == Line:
                # if so add to list to be returned
                grids_filtered.append(g)

    return grids_filtered


# --------------------------------------------- utility functions ------------------


def get_all_grid_heads_by_category(doc):
    """
    Gets all grid head types in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector with grid head types
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_GridHeads)
        .WhereElementIsElementType()
    )
    return collector


def get_all_grid_types_by_category(doc):
    """
    Gets all grid types in the model

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector with grid types
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_Grids)
        .WhereElementIsElementType()
    )
    return collector


def get_all_grid_type_ids_by_category(doc):
    """
    Gets all grid types ids in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector with grid type ids
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = get_all_grid_types_by_category(doc)
    ids = com.get_ids_from_element_collector(collector)
    return ids


def get_grid_type_names(doc, g):
    """
    Gets all valid grid types, based on a past in grid, available in model.

    Uses grid.GetValidTypes() to get the grid types.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param g: A grid
    :type g: Autodesk.Revit.DB.Grid
    :return: A nested set of lists containing grid type id and grid type name
    :rtype: list of lists [[GridTypeId as Revit ElementId, grid type name as string],[...]]
    """

    valid_grid_types = []
    valid_grid_type_ids = g.GetValidTypes()
    for valid_grid_type_id in valid_grid_type_ids:
        grid_data = []
        grid_type_t = doc.GetElement(valid_grid_type_id)
        grid_data.append(valid_grid_type_id)
        grid_data.append(Element.Name.GetValue(grid_type_t))
        valid_grid_types.append(grid_data)
    return valid_grid_types


def get_grid_type_name(doc, g):
    """
    Gets the grid type name of a grid.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param g: A grid.
    :type g: Autodesk.Revit.DB.Grid
    :return: The grid type name.
    :rtype: str
    """

    value = "unknown"
    grid_type_T = doc.GetElement(g.GetTypeId())
    value = Element.Name.GetValue(grid_type_T)
    return value


def get_grid_type_id_by_name(doc, grid_type_name):
    """
    Gets the grid type Id based on it's name, if no match found it returns the Revit Invalid Element Id

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param grid_type_name: The grid type name.
    :type grid_type_name: str
    :return: The grids type Id or if not match is found Autodesk.Revit.DB.ElementId.InvalidElementId
    :rtype: Autodesk.Revit.DB.ElementId
    """

    id = ElementId.InvalidElementId
    grids = FilteredElementCollector(doc).OfClass(Grid).ToList()
    if len(grids) > 0:
        g = grids[0]
        valid_grid_type_ids = g.GetValidTypes()
        for grid_typ_id in valid_grid_type_ids:
            g_type_name = Element.Name.GetValue(doc.GetElement(grid_typ_id))
            if g_type_name == grid_type_name:
                id = grid_typ_id
                break
    return id


def grid_check_parameter_value(g, para_name, para_condition, condition_value):
    """
    Returns true if a given parameter on a grid has a value meeting the parameter condition.

    :param g: A grid.
    :type g: Autodesk.Revit.DB.Grid
    :param para_name: A parameter Name.
    :type para_name: str
    :param para_condition: A function evaluating the parameter value. First argument is the value to be checked against. Second argument is the actual parameter value.
    :type para_condition: func(arg1,arg2)
    :param condition_value: The value to be checked against.
    :type condition_value: var
    :return: True if parameter value is evaluated to True otherwise False.
    :rtype: bool
    """

    rule_match = False
    p_value = rParaGet.get_parameter_value_by_name(g, para_name)
    if p_value != None:
        rule_match = rParaGet.check_parameter_value(g, para_condition, condition_value)
    return rule_match


def get_max_extent_as_string(g):
    """
    Gets the maximum extent of a grid.

    :param g: A grid.
    :type g: Autodesk.Revit.DB.Grid
    :return: A string in format [maxX,maxY,maxZ]
    :rtype: str
    """

    ex = g.GetExtents()
    max = get_point_as_string (ex.MaximumPoint)
    return max


def get_min_extent_as_string(g):
    """
    Gets the minimum extent of a grid.

    :param g: A grid.
    :type g: Autodesk.Revit.DB.Grid
    :return: A string in format [minX,minY,minZ]
    :rtype: str
    """

    ex = g.GetExtents()
    min =  get_point_as_string (ex.MinimumPoint)
    return min


def get_all_grid_head_family_type_ids(doc):
    """
    This will return all ids grid head family types in the model
    """

    ids = []
    filter = ElementCategoryFilter(BuiltInCategory.OST_GridHeads)
    col = FilteredElementCollector(doc).OfClass(FamilySymbol).WherePasses(filter)
    ids = com.get_ids_from_element_collector(col)
    return ids


def get_grid_curves_from_view(grid, view):
    """
    Gets grid curve (line from a view. Will return none if grid is out of view scope

    :param grid: The grid element.
    :type grid: Autodesk.Revit.DB.Grid
    :param view: The view of which the get the grid extend from.
    :type view: Autodesk.Revit.DB.View

    :return: A curve object if grid is visible in view, otherwise None
    :rtype: [Autodesk.Revit.DB.Curve] or None
    """

    try:
        curves = grid.GetCurvesInView(DatumExtentType.ViewSpecific, view)
        return curves
    except Exception as e:
        pass
    return None


def get_grid_plane_z_value(grids, view):
    """
    Get the z value of Grid plane by view based on first grid found visible in view.

    :param grids: Grids in the model
    :type grids: [Autodesk.Revit.DB.Grid]
    :param view: the view
    :type view: Autodesk.Revit.DB.View

    :return: The view plane Z value or 0.0 if no grid found in view.
    :rtype: float
    """

    z = 0.0
    try:
        for grid in grids:
            curve = get_grid_curves_from_view(grid, view)
            if curve != None:
                if len(curve) > 0:
                    z = curve[0].GetEndPoint(0).Z
                    break
    except Exception as e:
        pass
    return z



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit propagate grids extend visibility, and graphics to other views.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

from collections import namedtuple

from duHast.Revit.Grids import grids as rGrids
from duHast.Revit.Common import transaction as rTran
from duHast.Utilities.Objects import result as res
from duHast.Revit.Grids.grids import get_grid_curves_from_view, get_grid_plane_z_value
from duHast.UI.Objects.ProgressBase import ProgressBase

from Autodesk.Revit.DB import (
    DatumEnds,
    DatumExtentType,
    ElementId,
    Line,
    Transaction,
    XYZ,
)

# output indent
SPACER = "..."

# tuples containing grid data
grid_properties = namedtuple(
    "grid_properties",
    "is_grid_in_scope datum_extent_type_zero datum_extent_type_one grid_curve, end_zero_bubble, end_one_bubble, is_grid_hidden",
)


def get_active_view_grid_data(active_view, grids_in_model):
    """
    returns a dictionary where grid id is the key and values are a list of properties:

    - inScope (bool)
    - isVisible (bool)
    - end zero, end One: Extent type for both ends ( 2D vs 3D) (None if not in scope)
    - curve: the curve (line) describing the grid extent in view (None if not in scope)
    - end bubble one end bubble zero: bool describing which bubble isi visible in view (None if not in scope)

    :param active_view: The active view.
    :type active_view: Autodesk.Revit.DB.View
    :param grids_in_model: Grids in the model.
    :type grids_in_model: [Autodesk.Revit.DB.Grid]

    :return:
        Result class instance.

        - result.status: True.
        - result.message will contain the names of the grids not in scope only.
        - result.result contains a named tuple: grid_properties:

            {
                bool, is the grid in scope (inside the crop region)
                # Datum Extent type at End 0,
                # Datum Extend type at End 1,
                curve, describing the start and end point of grid
                bool, #bubble visible at zero end
                bool, # bubble visible at 1 end
                bool # is the grid visible
            }

        On exception

        - result.status (bool) will be True (exception is ignored)
        - result.message will contain the message grid not in scope for view

    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    grid_visibility_data = {}
    for grid in grids_in_model:
        try:
            # a grid can either be:
            # not in scope (outside the view crop zone)
            # hidden in view
            # visible in view
            grid_curve = get_grid_curves_from_view(grid=grid, view=active_view)
            # check whether grid is in scope (has a curve object for a given view)
            if grid_curve == None:
                grid_visibility_data[grid.Id] = grid_properties(
                    False, None, None, None, None, None, False
                )
            else:
                # store grid properties
                grid_visibility_data[grid.Id] = grid_properties(
                    True,
                    grid.GetDatumExtentTypeInView(DatumEnds.End0, active_view),
                    grid.GetDatumExtentTypeInView(DatumEnds.End1, active_view),
                    grid_curve,
                    grid.IsBubbleVisibleInView(DatumEnds.End0, active_view),
                    grid.IsBubbleVisibleInView(DatumEnds.End1, active_view),
                    grid.IsHidden(active_view),
                )
        except Exception as e:
            return_value.append_message(
                "{}{}Grid {} through exception in view: {} exception: {}. Setting grid to not in scope for view.".format(
                    SPACER, SPACER, grid.Name, active_view.Name, e
                )
            )
            grid_visibility_data[id] = grid_properties(
                False, None, None, None, None, None, False
            )
    return_value.result.append(grid_visibility_data)
    return return_value


def change_grid_extends_in_views(
    doc, grids_in_model, grid_template_data, views_to_change_grid_elements, callback
):
    """
    Changes grids  in a number views to match the extend, visibility, bubble visibility of grids in template view

    :param doc: Current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param grids_in_model: Grids in the model
    :type grids_in_model:  [Autodesk.Revit.DB.Grid]
    :param grid_template_data: A dictionary describing the appearance of each grid in the template view
    :type grid_template_data:  A named tuple: grid_properties
        bool, is the grid in scope (inside the crop region)
        # Datum Extent type at End 0,
        # Datum Extend type at End 1,
        curve, describing the start and end point of grid
        bool, # bubble visible at zero end
        bool, # bubble visible at 1 end
        bool # is the grid visible
    :param views_to_change_grid_elements: A list of views in which to update the grid graphics
    :type views_to_change_grid_elements: [Autodesk.Revit.DB.View]
    :param callback: A progress call back function
    :type views_to_change_grid_elements: :class:`.ProgressBase`
    """
    return_value = res.Result()

    # some type checking
    if isinstance(callback, ProgressBase) == False and callback is not None:
        return_value.update_sep(
            False,
            "Callback needs to inherit from ProgressBase or must be None. Got {} instead".format(
                type(callback)
            ),
        )
        return return_value

    try:
        view_counter = 1
        # loop over views and update grids based on template data
        for view in views_to_change_grid_elements:

            # update any progress call back
            if callback:
                callback.update(view_counter, len(views_to_change_grid_elements))

            return_value.append_message("Processing view: {}".format(view.Name))
            # get the grid z plane for the view
            new_grid_z_plane = get_grid_plane_z_value(grids_in_model, view)

            # loop over grid data
            for grid_id, item in grid_template_data.items():
                grid_element = doc.GetElement(grid_id)
                try:
                    # setup action to be performed in transaction
                    def action():
                        action_return_value = res.Result()
                        # check if grid is in scope of view
                        if item.is_grid_in_scope == True:
                            # check if grid vibility needs to be changed
                            if item.is_grid_hidden == True:
                                # check if grid is currently visible in view and needs to be hidden
                                if grid_element.IsHidden(view) == False:
                                    hide_ids = [grid_id]
                                    # hide the grid
                                    try:
                                        # need to convert into .net list
                                        view.HideElements(hide_ids.ToList[ElementId]())
                                        action_return_value.append_message(
                                            SPACER
                                            + "Hiding grid: {} in view: {}".format(
                                                grid_element.Name, view.Name
                                            )
                                        )
                                    except Exception as e:
                                        action_return_value.update_sep(
                                            False,
                                            "Hide grid, an exception ocurred: {}".format(
                                                e
                                            ),
                                        )
                                else:
                                    action_return_value.append_message(
                                        SPACER
                                        + "Grid: {} already hidden in view: {} No action required.".format(
                                            grid_element.Name, view.Name
                                        )
                                    )
                            else:
                                # check if grid is currently hidden in view and needs to be shown
                                try:
                                    if grid_element.IsHidden(view):
                                        un_hide_ids = [grid_id]
                                        view.UnhideElements(
                                            un_hide_ids.ToList[ElementId]()
                                        )
                                        action_return_value.append_message(
                                            SPACER
                                            + "Unhide grid: {} in view: {}".format(
                                                grid_element.Name, view.Name
                                            )
                                        )
                                except Exception as e:
                                    action_return_value.update_sep(
                                        False,
                                        "Unhide grid, an exception ocurred: {}".format(
                                            e
                                        ),
                                    )

                                # update other properties
                                action_return_value.append_message(
                                    SPACER
                                    + "Changing grid: {} properties in view: {}".format(
                                        grid_element.Name, view.Name
                                    )
                                )
                                # get the nested old curve from list
                                old_grid_curve = item.grid_curve[0]

                                # set extend types to match
                                grid_element.SetDatumExtentType(
                                    DatumEnds.End0, view, item.datum_extent_type_zero
                                )
                                grid_element.SetDatumExtentType(
                                    DatumEnds.End1, view, item.datum_extent_type_one
                                )

                                # set bubble visibility
                                if item.end_zero_bubble == True:
                                    grid_element.ShowBubbleInView(DatumEnds.End0, view)
                                else:
                                    grid_element.HideBubbleInView(DatumEnds.End0, view)
                                if item.end_one_bubble == True:
                                    grid_element.ShowBubbleInView(DatumEnds.End1, view)
                                else:
                                    grid_element.HideBubbleInView(DatumEnds.End1, view)

                                # set grid extend
                                start_point = XYZ(
                                    old_grid_curve.GetEndPoint(0).X,
                                    old_grid_curve.GetEndPoint(0).Y,
                                    new_grid_z_plane,
                                )
                                end_point = XYZ(
                                    old_grid_curve.GetEndPoint(1).X,
                                    old_grid_curve.GetEndPoint(1).Y,
                                    new_grid_z_plane,
                                )
                                new_grid_curve = Line.CreateBound(
                                    start_point, end_point
                                )

                                grid_element.SetCurveInView(
                                    DatumExtentType.ViewSpecific, view, new_grid_curve
                                )
                        else:
                            # grid is not in scope of this view
                            action_return_value.append_message(
                                SPACER
                                + "Grid: {} not in scope of view : {}".format(
                                    grid_element.Name, view.Name
                                )
                            )
                        return action_return_value

                    transaction = Transaction(doc, "Changing grid extends.")
                    result = rTran.in_transaction(transaction, action)
                    return_value.update(result)
                except Exception as e:
                    return_value.update_sep(
                        False,
                        "An exception ocurred when changing grid appearance: {}".format(
                            e
                        ),
                    )

            # check if cancelled
            if callback and callback.is_cancelled():
                return_value.append_message("User cancelled.")
                break
            
            # update the progress counter
            view_counter = view_counter + 1

    except Exception as e:
        return_value.update_sep(
            False, "An exception occurred while changing grid extends: {}".format(e)
        )

    return return_value


def propagate_grids_extends_and_visibility(
    doc, views_to_change_grid_elements, callback=None
):
    """
    Propagates grid extends and visibility from the active view to all other views selected in UI.

    :param doc: Current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param viewsToChangeGridElements: A list of views in which to update the grid graphics
    :type viewsToChangeGridElements: [Autodesk.Revit.DB.View]
    :param callback: A progress call back function
    :type views_to_change_grid_elements: :class:`.ProgressBase`

    :return:
        Result class instance.

        - result.status: True.
        - result.message will contain the names of the grids amended.
        - result.result an empty list

        On exception

        - result.status (bool) will be False
        - result.message will contain the exception message
    """

    return_value = res.Result()

    # some type checking
    if isinstance(callback, ProgressBase) == False and callback is not None:
        return_value.update_sep(
            False,
            "Callback needs to inherit from ProgressBase or must be None. Got {} instead".format(
                type(callback)
            ),
        )
        return return_value
    try:
        # get all grids in model
        grids_in_model = rGrids.get_grids_in_model(doc)
        if len(grids_in_model.ToElementIds()) > 0:
            # get active view grid information
            grid_data_result = get_active_view_grid_data(doc.ActiveView, grids_in_model)

            if grid_data_result.status == False:
                raise ValueError(grid_data_result.message)
            # get the grid template data
            grid_data = grid_data_result.result[0]

            # check if any views where provided
            if (
                views_to_change_grid_elements != None
                and len(views_to_change_grid_elements) > 0
            ):
                return_value.append_message(
                    "Amending grids in {} views:".format(
                        len(views_to_change_grid_elements)
                    )
                )

                # update grid extends
                result_change = change_grid_extends_in_views(
                    doc=doc,
                    grids_in_model=grids_in_model,
                    grid_template_data=grid_data,
                    views_to_change_grid_elements=views_to_change_grid_elements,
                    callback=callback,
                )
                return_value.update(result_change)
            else:
                return_value.append_message(
                    "No view to propagate grids extends to where provided."
                )
        else:
            return_value.append_message("No grids in model found.")
    except Exception as e:
        return_value.update_sep(
            False, "An exception occurred while propagating grid extends: {}".format(e)
        )
    return return_value



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import  ColorFillSchemeEntry, ElementId, StorageType

def get_entry_value_storage_type(entry):
    """
    Get the value and storage type of a given colour fill scheme entry.

    :param entry: The colour fill scheme entry from which to retrieve the value and storage type.
    :type entry: Autodesk.Revit.DB.ColorFillSchemeEntry
    
    :return: A tuple containing the value and storage type.
    :rtype: tuple
    """

    if isinstance(entry, ColorFillSchemeEntry) == False:
        raise TypeError("entry must be of type ColorFillSchemeEntry. Got {}".format(type(entry)))

    if entry.StorageType == StorageType.None:
        return (None,0)
    elif entry.StorageType == StorageType.Integer:
        return (entry.GetIntegerValue(), 1)
    elif entry.StorageType == StorageType.Double:
        return (entry.GetDoubleValue(),2)
    elif entry.StorageType == StorageType.String:
        return(entry.GetStringValue(),3)
    elif entry.StorageType == StorageType.ElementId:
       return( entry.GetElementIdValue(),4)


def get_entry_value_as_string(entry):
    """
    Get the value of a given colour fill scheme entry as a string.
    
    :param entry: The colour fill scheme entry from which to retrieve the value.
    :type entry: Autodesk.Revit.DB.ColorFillSchemeEntry
    :return: The value as a string.
    :rtype: str
    """

    value_type = get_entry_value_storage_type(entry)

    if value_type[0] == None: 
        return "None"
    elif value_type[1] == 1:
        return str(value_type[0])
    elif value_type[1] == 2:
        return str(value_type[0])
    elif value_type[1] == 3:
        return value_type[0]
    elif value_type[1] == 4:
        return str(value_type[0].IntegerValue)


def set_entry_value (entry, value):
    """
    Set the value of a given colour fill scheme entry depending on its storage type.
    
    :param entry: The colour fill scheme entry to set the value for.
    :type entry: Autodesk.Revit.DB.ColorFillSchemeEntry
    :param value: The value to set.
    :type value: str
    :return: None
    """
    
    if isinstance(entry, ColorFillSchemeEntry) == False:
        raise TypeError("entry must be of type ColorFillSchemeEntry. Got {}".format(type(entry)))

    if entry.StorageType == StorageType.None:
        raise TypeError("entry has no storage type. Got {}".format(type(entry)))
    elif entry.StorageType == StorageType.Integer:
        if isinstance(value, int):
            entry.SetIntegerValue(value)
        else:
            entry.SetIntegerValue(int(value))
    elif entry.StorageType == StorageType.Double:
        if isinstance(value, float):
            entry.SetDoubleValue(value)
        else:
            entry.SetDoubleValue(float(value))
    elif entry.StorageType ==  StorageType.String:
        if isinstance(value, str):
            entry.SetStringValue(value)
        else:
            entry.SetStringValue(str(value))
    elif entry.StorageType == StorageType.ElementId:
        if isinstance(value, ElementId):
            entry.SetElementIdValue(value)
        else:
            entry.SetElementIdValue(ElementId(int(value)))
    
    return entry


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A collection of common UI elements 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#
from Autodesk.Revit.UI import TaskDialog


class AlertDialog(TaskDialog):
    def __init__(self, content):
        self.Title = "BVN"
        TaskDialog.__init__(self.Title)
        self.MainContent = content


def alert(content, exitscript=True, additional_content=None):
    """
    This will display a task dialog with the given content. If exitscript
    is True, the script will terminate after the dialog is closed. Additional
    content input will be put in an expander. This could be more technical
    output
    :param content: The main content of the dialog
    :type content: str
    :param exitscript: If True, the script will terminate after dialog close, defaults to True
    :type exitscript: bool, optional
    :param additional_content: Additional content to be put in an expander, defaults to None
    :type additional_content: str, optional
    """
    if exitscript:
        dialog = AlertDialog(content)
        dialog.MainInstruction = "Error from script:"
        dialog.FooterText = "(Script will stop after this message)"
    else:
        dialog = AlertDialog(content)
        dialog.MainInstruction = "Notification:"

    # Turn off the Title prefix
    dialog.TitleAutoPrefix = False

    # Set additional content if provided
    if additional_content:
        dialog.ExpandedContent = additional_content

    # Show the dialog
    dialog.Show()

    # Exit script if required
    if exitscript:
        if additional_content:
            raise SystemExit(
                "Script terminated because {}\n\nFurther details:\n{}".format(
                    content, additional_content
                )
            )
        else:
            raise SystemExit("Script terminated because {}".format(content))



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a pyRevit file picker wrapper function. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

def get_file_path_from_user(forms, title, file_extension, multi_file=False):
    """
    Very simple wrapper function to get a file path from the user.

    :param forms: pyRevit forms
    :type forms: pyRevit forms module
    :param title: The title of the file picker dialog.
    :type title: str
    :param file_extension: The file extension to filter for. Example: "csv"
    :type file_extension: str
    :param multi_file: Single vs multi file selection, defaults to False (single file)
    :type multi_file: bool, optional

    :return: The file path(s) selected by the user. None if no file was selected.
    :rtype: str, [str],None
    """

    file_path = forms.pick_file(file_ext=file_extension, multi_file=multi_file, title=title)
    if file_path is None or file_path == "" or file_path == []:
        return None

    return file_path


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains the header row for any Revit sheet and views reports. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

#: header used in views report
from Autodesk.Revit.DB import (
    FilteredElementCollector,
    ViewPlan,
    ViewSheet,
    ViewSchedule,
)

#: header used in views report
REPORT_VIEWS_HEADER = ["HOSTFILE", "Id", "CREATED BY", "LAST CHANGED BY", "OWNER"]

#: header used in sheets report
REPORT_SHEETS_HEADER = ["HOSTFILE", "Id", "CREATED BY", "LAST CHANGED BY", "OWNER"]

#: header used in schedules report
REPORT_SCHEDULES_HEADER = ["HOSTFILE", "Id", "CREATED BY", "LAST CHANGED BY", "OWNER"]


def get_sheets_report_headers(doc):
    """
    A list of headers used in report files
    Hardcoded header list is expanded by parameters added to sheet category in model.

    Note: if no sheet views are in a file this will report the hardcoded header only.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of headers.
    :rtype: list str
    """

    collector_views = FilteredElementCollector(doc).OfClass(ViewSheet)
    # copy headers list
    headers = REPORT_SHEETS_HEADER[:]
    for v in collector_views:
        # get all parameters attached to sheet
        paras = v.GetOrderedParameters()
        for para in paras:
            headers.append(para.Definition.Name)
        break
    return headers


def get_schedules_report_headers(doc):
    """
    A list of headers used in report files
    Hardcoded header list is expanded by parameters added to PlanView class in model.

    Note: if no schedule views are in a file this will report the hardcoded header only.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of headers.
    :rtype: list str
    """

    collector_views = FilteredElementCollector(doc).OfClass(ViewSchedule)
    # copy headers list
    headers = REPORT_SCHEDULES_HEADER[:]
    for v in collector_views:
        # get all parameters attached to schedules
        paras = v.GetOrderedParameters()
        for para in paras:
            headers.append(para.Definition.Name)
        break
    return headers


def get_views_report_headers(doc):
    """
    A list of headers used in report files
    Hardcoded header list is expanded by parameters added to PlanView class in model.
    Revit ui only lists 'views' 'sheets' and 'schedules' as a categories to add project parameters to.
    Therefore any parameters attached to class ViewPlan should apply to all other view types with exception of schedules and sheets.

    Note: if no plan views are in a file this will report the hardcoded header only.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of headers.
    :rtype: list str
    """

    collector_views = FilteredElementCollector(doc).OfClass(ViewPlan)
    # copy headers list
    headers = REPORT_VIEWS_HEADER[:]
    for v in collector_views:
        # get all parameters attached to views
        paras = v.GetOrderedParameters()
        for para in paras:
            headers.append(para.Definition.Name)
        break
    return headers



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from System.Collections.Generic import List

from Autodesk.Revit.DB import  Document, ElementId, TableCellCombinedParameterData, ViewSheet

from duHast.Utilities.Objects import result as res
from duHast.UI.Objects.ProgressBase import ProgressBase
from duHast.Utilities.directory_io import directory_exists
from duHast.Revit.Exports.Utility.export_options_pdf_2024 import set_pdf_export_option_2024
from duHast.Revit.Exports.Utility.convert_pdf_dwg_settings import convert_settings_json_string_to_settings_objects

def create_naming_rule(sheet_name_string, sample_sheet):
    """
    Creates a naming rule for the PDF export based on the provided sheet name string.
    
    :param sheet_name_string: The sheet name string to use for the naming rule.
    :type sheet_name_string: str
    :return: The created naming rule.
    :rtype: str
    """
    

    # get all parameters from the sheet and id to build the naming rule
    para_dic = {}
    paras = sample_sheet.GetOrderedParameters()
    for p in paras:
        para_dic[p.Definition.Name] = p.Id

    # get settings from settings string
    settings =  convert_settings_json_string_to_settings_objects(sheet_name_string)
   
    # Create a list to hold the naming rules
    rules = List[TableCellCombinedParameterData]()

    # Build the naming rule using the provided string
    for setting in settings:

        # check if parameter exists if not skip this setting
        if setting.propertyName not in para_dic.keys():
            continue

        # Create a new naming rule
        rule = TableCellCombinedParameterData.Create()

        # set the prefix value if it is not None
        if setting.prefix is not None:
            rule.Prefix = setting.prefix
        
        if setting.suffix is not None:
            rule.Suffix = setting.suffix
        
        if setting.separator is not None:
            rule.Separator = setting.separator
        
        # check if this is a parameter name
        if setting.propertyName in para_dic.keys():
            rule.ParamId = para_dic[setting.propertyName]

        # add the rule to the list of rules 
        rules.Add(rule)

    return rules


def export_sheet_to_pdf (doc, view_sheet, sheet_name_string, output_directory):
    """
    Exports a Revit sheet to PDF using the provided naming rule and output directory.

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param view_sheet: The sheet to export.
    :type view_sheet: Autodesk.Revit.DB.ViewSheet
    :param sheet_name_string: The string to build the sheet name.
    :type sheet_name_string: str
    :param output_directory: The directory to save the PDF file.
    :type output_directory: str
    """

    return_value = res.Result()

    try:

        # Check if we got a document
        if not isinstance(doc, Document):
            raise TypeError("doc must be an instance of Autodesk.Revit.DB.Document")
    
        # Check if we got a list of sheets
        if not isinstance(view_sheet,  ViewSheet):
            raise TypeError("view_sheet must be an Autodesk.Revit.DB.ViewSheet instances")
        
        # Check if the output directory exists
        if not directory_exists(output_directory):
            return_value.update_sep(False, "Output directory does not exist.")
            return return_value

        # Check if the sheet name string is a string or None
        if not isinstance(sheet_name_string, str) and sheet_name_string is not None:
            raise TypeError("sheet_name_string must be a string or None")
        
        # Create the naming rule for the PDF export
        naming_rule = create_naming_rule(sheet_name_string, view_sheet)

        # Create the PDF export options
        pdf_export_option = set_pdf_export_option_2024(naming_rule)

        sheets = List[ElementId]()
        sheets.Add(view_sheet.Id)
        # Export the sheet to PDF
        export_result = doc.Export(output_directory, sheets,pdf_export_option)

        # reporting..
        sheet_identifier = "{} {}".format(view_sheet.SheetNumber, view_sheet.Name)

        if export_result:
            return_value.append_message("Sheet {} exported to PDF successfully.".format(sheet_identifier))
        else:
            return_value.update_sep(False, "Failed to export sheet {} to PDF.".format(sheet_identifier))

        return return_value
    
    except Exception as e:
            # handle the exception
            return_value.update_sep(False, "Error exporting sheets to PDF: {}".format(str(e)))
            return return_value


def export_sheets_to_pdf(doc, sheets, sheet_name_string, output_directory, callback=None):
    """
    Exports multiple Revit sheets to PDF using the provided naming rule and output directory.

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param sheets: The list of sheets to export.
    :type sheets: List[Autodesk.Revit.DB.ViewSheet]
    :param sheet_name_string: The string to build the sheet name.
    :type sheet_name_string: str
    :param output_directory: The directory to save the PDF files.
    :type output_directory: str
    """
    
    return_value = res.Result()

    try:

        # Check if we got a document
        if not isinstance(doc, Document):
            raise TypeError("doc must be an instance of Autodesk.Revit.DB.Document")
    
        # Check if we got a list of sheets
        if not isinstance(sheets, list) or not all(isinstance(sheet, ViewSheet) for sheet in sheets):
            raise TypeError("sheets must be a list of Autodesk.Revit.DB.ViewSheet instances")
        
        # Check if the output directory exists
        if not directory_exists(output_directory):
            return_value.update_sep(False, "Output directory does not exist.")
            return return_value

        # Check if the sheet name string is a string or None
        if not isinstance(sheet_name_string, str) and sheet_name_string is not None:
            raise TypeError("sheet_name_string must be a string or None")

        # check if the callback is of progressBase
        if callback is not None and not isinstance(callback, ProgressBase):
            raise TypeError("callback must be an instance of ProgressBase or None")


        # Check if the callback is None, if so export all in one go
        if callback is None:

            # Create the naming rule for the PDF export
            naming_rule = create_naming_rule(sheet_name_string, sheets[0])

            # Create the PDF export options
            pdf_export_option = set_pdf_export_option_2024(naming_rule=naming_rule)

            # convert to .net list
            sheets_net = List[ElementId]()
            for sheet in sheets:
                sheets_net.Add(sheet.Id)

            # Export the sheets to PDF
            export_result = doc.Export(output_directory, sheets_net, pdf_export_option)

            if export_result:
                return_value.update_sep(True, "Sheets exported to PDF successfully.")
            else:
                return_value.update_sep(False, "Failed to export sheets to PDF.")
        else:
            # export one sheet at the time and update the progress
            total_sheets = len(sheets)
            callback.update(0,  total_sheets)

            # loop over sheet at the time and export to pdf
            for i, sheet in enumerate(sheets):

                # update the progress
                callback.update(i, total_sheets)

                # Export the sheet to PDF
                export_result = export_sheet_to_pdf(doc, sheet, sheet_name_string, output_directory)
                return_value.update(export_result)

                # check if user cancel the export
                if callback.is_cancelled():
                    return_value.append_message("User cancelled!")
                    break
        
        return return_value
    except Exception as e:
        # handle the exception
        return_value.update_sep(False, "Error exporting sheets to PDF: {}".format(str(e)))
        return return_value
   


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Module containing reporting of family types in a project functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Reports:

- all family types and their type parameters values in a project

"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import Family

from duHast.Revit.Family.family_types_model_get_data_from_xml import (
    get_type_data_via_XML_from_family_object,
)
from duHast.Revit.Family.family_functions import get_name_and_category_to_family_dict
from duHast.Utilities.files_csv import read_csv_file
from duHast.Utilities.Objects.timer import Timer
from duHast.Utilities.Objects.result import Result
from duHast.UI.Objects.ProgressBase import ProgressBase


def build_type_report(doc_name, type_data, ignore_list_path):

    # list of differences
    report_data = []
    return_value = Result()
    try:

        # get ignore data
        ignore_data = []
        if ignore_list_path != None:
            ignore_data_result = read_csv_file(ignore_list_path)
            if ignore_data_result.status == False:
                return_value.update_sep(False, ignore_data_result.message)
                return return_value
            else:
                ignore_data = ignore_data_result.result

        # loop over the data and compare
        for entry in type_data:

            # check if entry is in ignore list
            ignore = False
            for ignore_entry in ignore_data:
                if (
                    entry.family_name == ignore_entry[0]
                    and entry.family_category == ignore_entry[1]
                ):
                    ignore = True
                    break

            # if not in ignore list, add to report
            if ignore:
                continue

            # add to report
            report_data = report_data + entry.get_report_data(doc_name)

    except Exception as e:
        return_value.update_sep(
            False, "Failed to build data report with exception: {}".format(e)
        )

    # store data to be returned
    return_value.result = report_data

    return return_value


def get_all_family_type_data_from_project_file(
    doc, ignore_list_path=None, progress_callback=None
):
    """
    Get all family type data from the project file.

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param ignore_list_path: The path to the ignore list file. ( 2 columns csv file with family name and category name)
    :type ignore_list_path: str
    :param progress_callback: The progress callback function.
    :type progress_callback: function
    :return: The family type data.
    :rtype: list[FamilyTypeDataStorageManager]
    """

    return_value = Result()

    # check callback class
    if progress_callback and isinstance(progress_callback, ProgressBase) == False:
        raise TypeError(
            "progress_callback needs to be inherited from ProgressBase. Got : {} instead.".format(
                type(progress_callback)
            )
        )

    # set up a timer
    t = Timer()
    t.start()

    data = []
    try:

        # get all families in the project
        families_loaded = get_name_and_category_to_family_dict(doc)

        # set progress counter
        counter = 1
        max_value_xml = len(families_loaded)

        # loop over loaded families and search for matches based on name and category
        for fam_name_and_category_name, revit_family in families_loaded.items():

            # update progress
            if progress_callback:
                progress_callback.update(counter, max_value_xml)

            # check this is a family
            if isinstance(revit_family, Family) is False:
                return_value.append_message(
                    "skipping family: {} as it is not a family".format(
                        fam_name_and_category_name
                    )
                )
                continue

            # ignore in place families
            if revit_family.IsInPlace:
                continue

            # create temp xml files from loaded family
            type_data_result = get_type_data_via_XML_from_family_object(
                revit_family=revit_family
            )
            if type_data_result.status == False:
                return_value.update_sep(
                    False,
                    "Failed to get type data from family: {} with exception: {}".format(
                        fam_name_and_category_name, type_data_result.message
                    ),
                )
                continue

            # get the type data
            data = data + type_data_result.result

            # update progress
            counter = counter + 1

            # check for user cancel
            if progress_callback != None:
                if progress_callback.is_cancelled():
                    return_value.append_message("User cancelled!")
                    break

        # build report from data
        return_value.append_message(
            "Successfully gathered family data: {}".format(t.stop())
        )

        # build report
        build_report_result = build_type_report(doc.Title, data, ignore_list_path)
        # check what came back
        return_value.update_sep(
            build_report_result.status, build_report_result.message
        )
        if build_report_result.status == False:
            return return_value
        # store the report data
        return_value.result = build_report_result.result

    except Exception as e:
        return_value.update_sep(
            False, "Failed to gather family data with exception: {}".format(e)
        )
        if t.is_running():
            return_value.append_message(t.stop())

    return return_value